#!/usr/bin/env python3
import os

import torch
from setuptools import setup
from torch.utils.cpp_extension import BuildExtension, CUDAExtension

cxx_args = ['-std=c++11']

nvcc_args = [
    '-gencode', 'arch=compute_52,code=sm_52',
    '-gencode', 'arch=compute_60,code=sm_60',
    '-gencode', 'arch=compute_61,code=sm_61',
    '-gencode', 'arch=compute_70,code=sm_70',
    '-gencode', 'arch=compute_70,code=compute_70'
]

setup(
    name='channelnorm_cuda',
    ext_modules=[
        CUDAExtension('channelnorm_cuda', [
            'channelnorm_cuda.cc',
            'channelnorm_kernel.cu'
        ], extra_compile_args={'cxx': cxx_args, 'nvcc': nvcc_args})
    ],
    cmdclass={
        'build_ext': BuildExtension
    })



import math

import numpy as np
import torch
import torch.nn as nn
from torch.nn import init

try:
    from networks import FlowNetC, FlowNetFusion, FlowNetS, FlowNetSD
    from networks.channelnorm_package.channelnorm import ChannelNorm
    from networks.resample2d_package.resample2d import Resample2d
    from networks.submodules import *
except:
    from .networks import FlowNetC, FlowNetFusion, FlowNetS, FlowNetSD
    from .networks.channelnorm_package.channelnorm import ChannelNorm
    from .networks.resample2d_package.resample2d import Resample2d
    from .networks.submodules import *
'Parameter count = 162,518,834'

class FlowNet2(nn.Module):

    def __init__(self, args, batchNorm=False, div_flow = 20.):
        super(FlowNet2,self).__init__()
        self.batchNorm = batchNorm
        self.div_flow = div_flow
        self.rgb_max = args.rgb_max
        self.args = args

        self.channelnorm = ChannelNorm()

        # First Block (FlowNetC)
        self.flownetc = FlowNetC.FlowNetC(args, batchNorm=self.batchNorm)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear')

        if args.fp16:
            self.resample1 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample1 = Resample2d()

        # Block (FlowNetS1)
        self.flownets_1 = FlowNetS.FlowNetS(args, batchNorm=self.batchNorm)
        self.upsample2 = nn.Upsample(scale_factor=4, mode='bilinear')
        if args.fp16:
            self.resample2 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample2 = Resample2d()


        # Block (FlowNetS2)
        self.flownets_2 = FlowNetS.FlowNetS(args, batchNorm=self.batchNorm)

        # Block (FlowNetSD)
        self.flownets_d = FlowNetSD.FlowNetSD(args, batchNorm=self.batchNorm) 
        self.upsample3 = nn.Upsample(scale_factor=4, mode='nearest') 
        self.upsample4 = nn.Upsample(scale_factor=4, mode='nearest') 

        if args.fp16:
            self.resample3 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample3 = Resample2d()

        if args.fp16:
            self.resample4 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample4 = Resample2d()

        # Block (FLowNetFusion)
        self.flownetfusion = FlowNetFusion.FlowNetFusion(args, batchNorm=self.batchNorm)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform_(m.bias)
                init.xavier_uniform_(m.weight)
                # init_deconv_bilinear(m.weight)

    def init_deconv_bilinear(self, weight):
        f_shape = weight.size()
        heigh, width = f_shape[-2], f_shape[-1]
        f = np.ceil(width/2.0)
        c = (2 * f - 1 - f % 2) / (2.0 * f)
        bilinear = np.zeros([heigh, width])
        for x in range(width):
            for y in range(heigh):
                value = (1 - abs(x / f - c)) * (1 - abs(y / f - c))
                bilinear[x, y] = value
        min_dim = min(f_shape[0], f_shape[1])
        weight.data.fill_(0.)
        for i in range(min_dim):
            weight.data[i,i,:,:] = torch.from_numpy(bilinear)
        return 

    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        
        x = (inputs - rgb_mean) / self.rgb_max
        x1 = x[:,:,0,:,:]
        x2 = x[:,:,1,:,:]
        x = torch.cat((x1,x2), dim = 1)

        # flownetc
        flownetc_flow2 = self.flownetc(x)[0]
        flownetc_flow = self.upsample1(flownetc_flow2*self.div_flow)
        
        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.resample1(x[:,3:,:,:], flownetc_flow)
        diff_img0 = x[:,:3,:,:] - resampled_img1 
        norm_diff_img0 = self.channelnorm(diff_img0)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((x, resampled_img1, flownetc_flow/self.div_flow, norm_diff_img0), dim=1)
        
        # flownets1
        flownets1_flow2 = self.flownets_1(concat1)[0]
        flownets1_flow = self.upsample2(flownets1_flow2*self.div_flow) 

        # warp img1 to img0 using flownets1; magnitude of diff between img0 and and warped_img1
        resampled_img1 = self.resample2(x[:,3:,:,:], flownets1_flow)
        diff_img0 = x[:,:3,:,:] - resampled_img1
        norm_diff_img0 = self.channelnorm(diff_img0)

        # concat img0, img1, img1->img0, flow, diff-mag
        concat2 = torch.cat((x, resampled_img1, flownets1_flow/self.div_flow, norm_diff_img0), dim=1)

        # flownets2
        flownets2_flow2 = self.flownets_2(concat2)[0]
        flownets2_flow = self.upsample4(flownets2_flow2 * self.div_flow)
        norm_flownets2_flow = self.channelnorm(flownets2_flow)

        diff_flownets2_flow = self.resample4(x[:,3:,:,:], flownets2_flow)
        # if not diff_flownets2_flow.volatile:
        #     diff_flownets2_flow.register_hook(save_grad(self.args.grads, 'diff_flownets2_flow'))

        diff_flownets2_img1 = self.channelnorm((x[:,:3,:,:]-diff_flownets2_flow))
        # if not diff_flownets2_img1.volatile:
        #     diff_flownets2_img1.register_hook(save_grad(self.args.grads, 'diff_flownets2_img1'))

        # flownetsd
        flownetsd_flow2 = self.flownets_d(x)[0]
        flownetsd_flow = self.upsample3(flownetsd_flow2 / self.div_flow)
        norm_flownetsd_flow = self.channelnorm(flownetsd_flow)
        
        diff_flownetsd_flow = self.resample3(x[:,3:,:,:], flownetsd_flow)
        # if not diff_flownetsd_flow.volatile:
        #     diff_flownetsd_flow.register_hook(save_grad(self.args.grads, 'diff_flownetsd_flow'))

        diff_flownetsd_img1 = self.channelnorm((x[:,:3,:,:]-diff_flownetsd_flow))
        # if not diff_flownetsd_img1.volatile:
        #     diff_flownetsd_img1.register_hook(save_grad(self.args.grads, 'diff_flownetsd_img1'))

        # concat img1 flownetsd, flownets2, norm_flownetsd, norm_flownets2, diff_flownetsd_img1, diff_flownets2_img1
        concat3 = torch.cat((x[:,:3,:,:], flownetsd_flow, flownets2_flow, norm_flownetsd_flow, norm_flownets2_flow, diff_flownetsd_img1, diff_flownets2_img1), dim=1)
        flownetfusion_flow = self.flownetfusion(concat3)

        # if not flownetfusion_flow.volatile:
        #     flownetfusion_flow.register_hook(save_grad(self.args.grads, 'flownetfusion_flow'))

        return flownetfusion_flow

class FlowNet2C(FlowNetC.FlowNetC):
    def __init__(self, args, batchNorm=False, div_flow=20):
        super(FlowNet2C,self).__init__(args, batchNorm=batchNorm, div_flow=20)
        self.rgb_max = args.rgb_max

    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        
        x = (inputs - rgb_mean) / self.rgb_max
        x1 = x[:,:,0,:,:]
        x2 = x[:,:,1,:,:]

        # FlownetC top input stream
        out_conv1a = self.conv1(x1)
        out_conv2a = self.conv2(out_conv1a)
        out_conv3a = self.conv3(out_conv2a)

        # FlownetC bottom input stream
        out_conv1b = self.conv1(x2)
        
        out_conv2b = self.conv2(out_conv1b)
        out_conv3b = self.conv3(out_conv2b)

        # Merge streams
        out_corr = self.corr(out_conv3a, out_conv3b) # False
        out_corr = self.corr_activation(out_corr)

        # Redirect top input stream and concatenate
        out_conv_redir = self.conv_redir(out_conv3a)

        in_conv3_1 = torch.cat((out_conv_redir, out_corr), 1)

        # Merged conv layers
        out_conv3_1 = self.conv3_1(in_conv3_1)

        out_conv4 = self.conv4_1(self.conv4(out_conv3_1))

        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)

        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)

        flow5       = self.predict_flow5(concat5)
        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)

        flow4       = self.predict_flow4(concat4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        concat3 = torch.cat((out_conv3_1,out_deconv3,flow4_up),1)

        flow3       = self.predict_flow3(concat3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)
        concat2 = torch.cat((out_conv2a,out_deconv2,flow3_up),1)

        flow2 = self.predict_flow2(concat2)

        if self.training:
            return flow2,flow3,flow4,flow5,flow6
        else:
            return self.upsample1(flow2*self.div_flow)

class FlowNet2S(FlowNetS.FlowNetS):
    def __init__(self, args, batchNorm=False, div_flow=20):
        super(FlowNet2S,self).__init__(args, input_channels = 6, batchNorm=batchNorm)
        self.rgb_max = args.rgb_max
        self.div_flow = div_flow
        
    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        x = (inputs - rgb_mean) / self.rgb_max
        x = torch.cat( (x[:,:,0,:,:], x[:,:,1,:,:]), dim = 1)

        out_conv1 = self.conv1(x)

        out_conv2 = self.conv2(out_conv1)
        out_conv3 = self.conv3_1(self.conv3(out_conv2))
        out_conv4 = self.conv4_1(self.conv4(out_conv3))
        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)
        
        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)
        flow5       = self.predict_flow5(concat5)
        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)
        flow4       = self.predict_flow4(concat4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        
        concat3 = torch.cat((out_conv3,out_deconv3,flow4_up),1)
        flow3       = self.predict_flow3(concat3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)

        concat2 = torch.cat((out_conv2,out_deconv2,flow3_up),1)
        flow2 = self.predict_flow2(concat2)

        if self.training:
            return flow2,flow3,flow4,flow5,flow6
        else:
            return self.upsample1(flow2*self.div_flow)

class FlowNet2SD(FlowNetSD.FlowNetSD):
    def __init__(self, args, batchNorm=False, div_flow=20):
        super(FlowNet2SD,self).__init__(args, batchNorm=batchNorm)
        self.rgb_max = args.rgb_max
        self.div_flow = div_flow

    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        x = (inputs - rgb_mean) / self.rgb_max
        x = torch.cat( (x[:,:,0,:,:], x[:,:,1,:,:]), dim = 1)

        out_conv0 = self.conv0(x)
        out_conv1 = self.conv1_1(self.conv1(out_conv0))
        out_conv2 = self.conv2_1(self.conv2(out_conv1))

        out_conv3 = self.conv3_1(self.conv3(out_conv2))
        out_conv4 = self.conv4_1(self.conv4(out_conv3))
        out_conv5 = self.conv5_1(self.conv5(out_conv4))
        out_conv6 = self.conv6_1(self.conv6(out_conv5))

        flow6       = self.predict_flow6(out_conv6)
        flow6_up    = self.upsampled_flow6_to_5(flow6)
        out_deconv5 = self.deconv5(out_conv6)
        
        concat5 = torch.cat((out_conv5,out_deconv5,flow6_up),1)
        out_interconv5 = self.inter_conv5(concat5)
        flow5       = self.predict_flow5(out_interconv5)

        flow5_up    = self.upsampled_flow5_to_4(flow5)
        out_deconv4 = self.deconv4(concat5)
        
        concat4 = torch.cat((out_conv4,out_deconv4,flow5_up),1)
        out_interconv4 = self.inter_conv4(concat4)
        flow4       = self.predict_flow4(out_interconv4)
        flow4_up    = self.upsampled_flow4_to_3(flow4)
        out_deconv3 = self.deconv3(concat4)
        
        concat3 = torch.cat((out_conv3,out_deconv3,flow4_up),1)
        out_interconv3 = self.inter_conv3(concat3)
        flow3       = self.predict_flow3(out_interconv3)
        flow3_up    = self.upsampled_flow3_to_2(flow3)
        out_deconv2 = self.deconv2(concat3)

        concat2 = torch.cat((out_conv2,out_deconv2,flow3_up),1)
        out_interconv2 = self.inter_conv2(concat2)
        flow2 = self.predict_flow2(out_interconv2)

        if self.training:
            return flow2,flow3,flow4,flow5,flow6
        else:
            return self.upsample1(flow2*self.div_flow)

class FlowNet2CS(nn.Module):

    def __init__(self, args, batchNorm=False, div_flow = 20.):
        super(FlowNet2CS,self).__init__()
        self.batchNorm = batchNorm
        self.div_flow = div_flow
        self.rgb_max = args.rgb_max
        self.args = args

        self.channelnorm = ChannelNorm()

        # First Block (FlowNetC)
        self.flownetc = FlowNetC.FlowNetC(args, batchNorm=self.batchNorm)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear')

        if args.fp16:
            self.resample1 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample1 = Resample2d()

        # Block (FlowNetS1)
        self.flownets_1 = FlowNetS.FlowNetS(args, batchNorm=self.batchNorm)
        self.upsample2 = nn.Upsample(scale_factor=4, mode='bilinear')

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform(m.bias)
                init.xavier_uniform(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform(m.bias)
                init.xavier_uniform(m.weight)
                # init_deconv_bilinear(m.weight)

    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        
        x = (inputs - rgb_mean) / self.rgb_max
        x1 = x[:,:,0,:,:]
        x2 = x[:,:,1,:,:]
        x = torch.cat((x1,x2), dim = 1)

        # flownetc
        flownetc_flow2 = self.flownetc(x)[0]
        flownetc_flow = self.upsample1(flownetc_flow2*self.div_flow)
        
        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.resample1(x[:,3:,:,:], flownetc_flow)
        diff_img0 = x[:,:3,:,:] - resampled_img1 
        norm_diff_img0 = self.channelnorm(diff_img0)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((x, resampled_img1, flownetc_flow/self.div_flow, norm_diff_img0), dim=1)
        
        # flownets1
        flownets1_flow2 = self.flownets_1(concat1)[0]
        flownets1_flow = self.upsample2(flownets1_flow2*self.div_flow) 

        return flownets1_flow

class FlowNet2CSS(nn.Module):

    def __init__(self, args, batchNorm=False, div_flow = 20.):
        super(FlowNet2CSS,self).__init__()
        self.batchNorm = batchNorm
        self.div_flow = div_flow
        self.rgb_max = args.rgb_max
        self.args = args

        self.channelnorm = ChannelNorm()

        # First Block (FlowNetC)
        self.flownetc = FlowNetC.FlowNetC(args, batchNorm=self.batchNorm)
        self.upsample1 = nn.Upsample(scale_factor=4, mode='bilinear')

        if args.fp16:
            self.resample1 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample1 = Resample2d()

        # Block (FlowNetS1)
        self.flownets_1 = FlowNetS.FlowNetS(args, batchNorm=self.batchNorm)
        self.upsample2 = nn.Upsample(scale_factor=4, mode='bilinear')
        if args.fp16:
            self.resample2 = nn.Sequential(
                            tofp32(), 
                            Resample2d(),
                            tofp16()) 
        else:
            self.resample2 = Resample2d()


        # Block (FlowNetS2)
        self.flownets_2 = FlowNetS.FlowNetS(args, batchNorm=self.batchNorm)
        self.upsample3 = nn.Upsample(scale_factor=4, mode='nearest') 

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                if m.bias is not None:
                    init.uniform(m.bias)
                init.xavier_uniform(m.weight)

            if isinstance(m, nn.ConvTranspose2d):
                if m.bias is not None:
                    init.uniform(m.bias)
                init.xavier_uniform(m.weight)
                # init_deconv_bilinear(m.weight)

    def forward(self, inputs):
        rgb_mean = inputs.contiguous().view(inputs.size()[:2]+(-1,)).mean(dim=-1).view(inputs.size()[:2] + (1,1,1,))
        
        x = (inputs - rgb_mean) / self.rgb_max
        x1 = x[:,:,0,:,:]
        x2 = x[:,:,1,:,:]
        x = torch.cat((x1,x2), dim = 1)

        # flownetc
        flownetc_flow2 = self.flownetc(x)[0]
        flownetc_flow = self.upsample1(flownetc_flow2*self.div_flow)
        
        # warp img1 to img0; magnitude of diff between img0 and and warped_img1, 
        resampled_img1 = self.resample1(x[:,3:,:,:], flownetc_flow)
        diff_img0 = x[:,:3,:,:] - resampled_img1 
        norm_diff_img0 = self.channelnorm(diff_img0)

        # concat img0, img1, img1->img0, flow, diff-mag ; 
        concat1 = torch.cat((x, resampled_img1, flownetc_flow/self.div_flow, norm_diff_img0), dim=1)
        
        # flownets1
        flownets1_flow2 = self.flownets_1(concat1)[0]
        flownets1_flow = self.upsample2(flownets1_flow2*self.div_flow) 

        # warp img1 to img0 using flownets1; magnitude of diff between img0 and and warped_img1
        resampled_img1 = self.resample2(x[:,3:,:,:], flownets1_flow)
        diff_img0 = x[:,:3,:,:] - resampled_img1
        norm_diff_img0 = self.channelnorm(diff_img0)

        # concat img0, img1, img1->img0, flow, diff-mag
        concat2 = torch.cat((x, resampled_img1, flownets1_flow/self.div_flow, norm_diff_img0), dim=1)

        # flownets2
        flownets2_flow2 = self.flownets_2(concat2)[0]
        flownets2_flow = self.upsample3(flownets2_flow2 * self.div_flow)

        return flownets2_flow




# freda (todo) : 

import numpy as np
import torch
import torch.nn as nn


def conv(batchNorm, in_planes, out_planes, kernel_size=3, stride=1):
    if batchNorm:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=(kernel_size-1)//2, bias=False),
            nn.BatchNorm2d(out_planes),
            nn.LeakyReLU(0.1,inplace=True)
        )
    else:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=(kernel_size-1)//2, bias=True),
            nn.LeakyReLU(0.1,inplace=True)
        )

def i_conv(batchNorm, in_planes, out_planes, kernel_size=3, stride=1, bias = True):
    if batchNorm:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=(kernel_size-1)//2, bias=bias),
            nn.BatchNorm2d(out_planes),
        )
    else:
        return nn.Sequential(
            nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=(kernel_size-1)//2, bias=bias),
        )

def predict_flow(in_planes):
    return nn.Conv2d(in_planes,2,kernel_size=3,stride=1,padding=1,bias=True)

def deconv(in_planes, out_planes):
    return nn.Sequential(
        nn.ConvTranspose2d(in_planes, out_planes, kernel_size=4, stride=2, padding=1, bias=True),
        nn.LeakyReLU(0.1,inplace=True)
    )

class tofp16(nn.Module):
    def __init__(self):
        super(tofp16, self).__init__()

    def forward(self, input):
        return input.half()


class tofp32(nn.Module):
    def __init__(self):
        super(tofp32, self).__init__()

    def forward(self, input):
        return input.float()


def init_deconv_bilinear(weight):
    f_shape = weight.size()
    heigh, width = f_shape[-2], f_shape[-1]
    f = np.ceil(width/2.0)
    c = (2 * f - 1 - f % 2) / (2.0 * f)
    bilinear = np.zeros([heigh, width])
    for x in range(width):
        for y in range(heigh):
            value = (1 - abs(x / f - c)) * (1 - abs(y / f - c))
            bilinear[x, y] = value
    weight.data.fill_(0.)
    for i in range(f_shape[0]):
        for j in range(f_shape[1]):
            weight.data[i,j,:,:] = torch.from_numpy(bilinear)


def save_grad(grads, name):
    def hook(grad):
        grads[name] = grad
    return hook

'''
def save_grad(grads, name):
    def hook(grad):
        grads[name] = grad
    return hook
import torch
from channelnorm_package.modules.channelnorm import ChannelNorm 
model = ChannelNorm().cuda()
grads = {}
a = 100*torch.autograd.Variable(torch.randn((1,3,5,5)).cuda(), requires_grad=True)
a.register_hook(save_grad(grads, 'a'))
b = model(a)
y = torch.mean(b)
y.backward()

'''



import os.path as osp
import subprocess

import skvideo.io


def get_video_duration(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError("{} not exists".format(video_path))
    metadata = skvideo.io.ffprobe(video_path)
    if '@duration' not in metadata['video']:
        result = subprocess.Popen(
            ["ffprobe", video_path],
            stdout=subprocess.PIPE, stderr = subprocess.STDOUT)
        duration = [x.decode('utf-8') for x in result.stdout.readlines()
                    if "Duration" in x.decode('utf-8')]
        duration_in_second = float(
            duration[0][15:17]) * 60 + float(duration[0][18:23])
        return duration_in_second
    return float(metadata['video']['@duration'])


def get_video_count(video_path):
    video_path = str(video_path)
    if not osp.exists(video_path):
        raise OSError("{} not exists".format(video_path))
    metadata = skvideo.io.ffprobe(video_path)
    if '@nb_frames' not in metadata['video']:
        nb_frames = subprocess.Popen(
            'ffprobe -v error -count_frames -select_streams v:0 -show_entries stream=nb_read_frames -of default=nokey=1:noprint_wrappers=1 {}'.format(video_path).split(),
            stdout=subprocess.PIPE, stderr = subprocess.STDOUT).stdout.read()
        return int(nb_frames)
    return int(metadata['video']['@nb_frames'])



from torch.autograd import Function, Variable
from torch.nn.modules.module import Module

import channelnorm_cuda


class ChannelNormFunction(Function):

    @staticmethod
    def forward(ctx, input1, norm_deg=2):
        assert input1.is_contiguous()
        b, _, h, w = input1.size()
        output = input1.new(b, 1, h, w).zero_()

        channelnorm_cuda.forward(input1, output, norm_deg)
        ctx.save_for_backward(input1, output)
        ctx.norm_deg = norm_deg

        return output

    @staticmethod
    def backward(ctx, grad_output):
        input1, output = ctx.saved_tensors

        grad_input1 = Variable(input1.new(input1.size()).zero_())

        channelnorm_cuda.backward(input1, output, grad_output.data,
                                              grad_input1.data, ctx.norm_deg)

        return grad_input1, None


class ChannelNorm(Module):

    def __init__(self, norm_deg=2):
        super(ChannelNorm, self).__init__()
        self.norm_deg = norm_deg

    def forward(self, input1):
        return ChannelNormFunction.apply(input1, self.norm_deg)




# flake8: noqa

import torch
import torch.nn as nn
import torch.nn.functional as F

from anet_tools.features.data import get_bn_inception_pretrained_model


class BNInception(nn.Module):

    def __init__(self, pretrained_model=True):
        super(BNInception, self).__init__()
        self.conv1_7x7_s2 = nn.Conv2d(
            10, 64, kernel_size=(
                7, 7), stride=(
                2, 2), padding=(
                3, 3), dilation=1, groups=1, bias=True)
        self.conv1_7x7_s2_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.conv1_relu_7x7 = nn.ReLU()
        self.pool1_3x3_s2 = nn.MaxPool2d(
            kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                0, 0), ceil_mode=True)
        self.conv2_3x3_reduce = nn.Conv2d(
            64, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.conv2_3x3_reduce_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.conv2_relu_3x3_reduce = nn.ReLU()
        self.conv2_3x3 = nn.Conv2d(
            64, 192, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.conv2_3x3_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.conv2_relu_3x3 = nn.ReLU()
        self.pool2_3x3_s2 = nn.MaxPool2d(
            kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                0, 0), ceil_mode=True)
        self.inception_3a_1x1 = nn.Conv2d(
            192, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3a_1x1_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3a_relu_1x1 = nn.ReLU()
        self.inception_3a_3x3_reduce = nn.Conv2d(
            192, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3a_3x3_reduce_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3a_relu_3x3_reduce = nn.ReLU()
        self.inception_3a_3x3 = nn.Conv2d(
            64, 64, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3a_3x3_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3a_relu_3x3 = nn.ReLU()
        self.inception_3a_double_3x3_reduce = nn.Conv2d(
            192, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3a_double_3x3_reduce_bn = nn.BatchNorm2d(
            64, momentum=0.1)
        self.inception_3a_relu_double_3x3_reduce = nn.ReLU()
        self.inception_3a_double_3x3_1 = nn.Conv2d(
            64, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3a_double_3x3_1_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3a_relu_double_3x3_1 = nn.ReLU()
        self.inception_3a_double_3x3_2 = nn.Conv2d(
            96, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3a_double_3x3_2_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3a_relu_double_3x3_2 = nn.ReLU()
        self.inception_3a_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_3a_pool_proj = nn.Conv2d(
            192, 32, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3a_pool_proj_bn = nn.BatchNorm2d(32, momentum=0.1)
        self.inception_3a_relu_pool_proj = nn.ReLU()
        self.inception_3b_1x1 = nn.Conv2d(
            256, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3b_1x1_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3b_relu_1x1 = nn.ReLU()
        self.inception_3b_3x3_reduce = nn.Conv2d(
            256, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3b_3x3_reduce_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3b_relu_3x3_reduce = nn.ReLU()
        self.inception_3b_3x3 = nn.Conv2d(
            64, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3b_3x3_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3b_relu_3x3 = nn.ReLU()
        self.inception_3b_double_3x3_reduce = nn.Conv2d(
            256, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3b_double_3x3_reduce_bn = nn.BatchNorm2d(
            64, momentum=0.1)
        self.inception_3b_relu_double_3x3_reduce = nn.ReLU()
        self.inception_3b_double_3x3_1 = nn.Conv2d(
            64, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3b_double_3x3_1_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3b_relu_double_3x3_1 = nn.ReLU()
        self.inception_3b_double_3x3_2 = nn.Conv2d(
            96, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3b_double_3x3_2_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3b_relu_double_3x3_2 = nn.ReLU()
        self.inception_3b_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_3b_pool_proj = nn.Conv2d(
            256, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3b_pool_proj_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_3b_relu_pool_proj = nn.ReLU()
        self.inception_3c_3x3_reduce = nn.Conv2d(
            320, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3c_3x3_reduce_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_3c_relu_3x3_reduce = nn.ReLU()
        self.inception_3c_3x3 = nn.Conv2d(
            128, 160, kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3c_3x3_bn = nn.BatchNorm2d(160, momentum=0.1)
        self.inception_3c_relu_3x3 = nn.ReLU()
        self.inception_3c_double_3x3_reduce = nn.Conv2d(
            320, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_3c_double_3x3_reduce_bn = nn.BatchNorm2d(
            64, momentum=0.1)
        self.inception_3c_relu_double_3x3_reduce = nn.ReLU()
        self.inception_3c_double_3x3_1 = nn.Conv2d(
            64, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3c_double_3x3_1_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3c_relu_double_3x3_1 = nn.ReLU()
        self.inception_3c_double_3x3_2 = nn.Conv2d(
            96, 96, kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_3c_double_3x3_2_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_3c_relu_double_3x3_2 = nn.ReLU()
        self.inception_3c_pool = nn.MaxPool2d(
            kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                0, 0), ceil_mode=True)
        self.inception_4a_1x1 = nn.Conv2d(
            576, 224, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4a_1x1_bn = nn.BatchNorm2d(224, momentum=0.1)
        self.inception_4a_relu_1x1 = nn.ReLU()
        self.inception_4a_3x3_reduce = nn.Conv2d(
            576, 64, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4a_3x3_reduce_bn = nn.BatchNorm2d(64, momentum=0.1)
        self.inception_4a_relu_3x3_reduce = nn.ReLU()
        self.inception_4a_3x3 = nn.Conv2d(
            64, 96, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4a_3x3_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_4a_relu_3x3 = nn.ReLU()
        self.inception_4a_double_3x3_reduce = nn.Conv2d(
            576, 96, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4a_double_3x3_reduce_bn = nn.BatchNorm2d(
            96, momentum=0.1)
        self.inception_4a_relu_double_3x3_reduce = nn.ReLU()
        self.inception_4a_double_3x3_1 = nn.Conv2d(
            96, 128, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4a_double_3x3_1_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4a_relu_double_3x3_1 = nn.ReLU()
        self.inception_4a_double_3x3_2 = nn.Conv2d(
            128, 128, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4a_double_3x3_2_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4a_relu_double_3x3_2 = nn.ReLU()
        self.inception_4a_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_4a_pool_proj = nn.Conv2d(
            576, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4a_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4a_relu_pool_proj = nn.ReLU()
        self.inception_4b_1x1 = nn.Conv2d(
            576, 192, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4b_1x1_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_4b_relu_1x1 = nn.ReLU()
        self.inception_4b_3x3_reduce = nn.Conv2d(
            576, 96, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4b_3x3_reduce_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_4b_relu_3x3_reduce = nn.ReLU()
        self.inception_4b_3x3 = nn.Conv2d(
            96, 128, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4b_3x3_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4b_relu_3x3 = nn.ReLU()
        self.inception_4b_double_3x3_reduce = nn.Conv2d(
            576, 96, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4b_double_3x3_reduce_bn = nn.BatchNorm2d(
            96, momentum=0.1)
        self.inception_4b_relu_double_3x3_reduce = nn.ReLU()
        self.inception_4b_double_3x3_1 = nn.Conv2d(
            96, 128, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4b_double_3x3_1_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4b_relu_double_3x3_1 = nn.ReLU()
        self.inception_4b_double_3x3_2 = nn.Conv2d(
            128, 128, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4b_double_3x3_2_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4b_relu_double_3x3_2 = nn.ReLU()
        self.inception_4b_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_4b_pool_proj = nn.Conv2d(
            576, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4b_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4b_relu_pool_proj = nn.ReLU()
        self.inception_4c_1x1 = nn.Conv2d(
            576, 160, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4c_1x1_bn = nn.BatchNorm2d(160, momentum=0.1)
        self.inception_4c_relu_1x1 = nn.ReLU()
        self.inception_4c_3x3_reduce = nn.Conv2d(
            576, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4c_3x3_reduce_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4c_relu_3x3_reduce = nn.ReLU()
        self.inception_4c_3x3 = nn.Conv2d(
            128, 160, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4c_3x3_bn = nn.BatchNorm2d(160, momentum=0.1)
        self.inception_4c_relu_3x3 = nn.ReLU()
        self.inception_4c_double_3x3_reduce = nn.Conv2d(
            576, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4c_double_3x3_reduce_bn = nn.BatchNorm2d(
            128, momentum=0.1)
        self.inception_4c_relu_double_3x3_reduce = nn.ReLU()
        self.inception_4c_double_3x3_1 = nn.Conv2d(
            128, 160, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4c_double_3x3_1_bn = nn.BatchNorm2d(160, momentum=0.1)
        self.inception_4c_relu_double_3x3_1 = nn.ReLU()
        self.inception_4c_double_3x3_2 = nn.Conv2d(
            160, 160, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4c_double_3x3_2_bn = nn.BatchNorm2d(160, momentum=0.1)
        self.inception_4c_relu_double_3x3_2 = nn.ReLU()
        self.inception_4c_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_4c_pool_proj = nn.Conv2d(
            576, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4c_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4c_relu_pool_proj = nn.ReLU()
        self.inception_4d_1x1 = nn.Conv2d(
            608, 96, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4d_1x1_bn = nn.BatchNorm2d(96, momentum=0.1)
        self.inception_4d_relu_1x1 = nn.ReLU()
        self.inception_4d_3x3_reduce = nn.Conv2d(
            608, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4d_3x3_reduce_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4d_relu_3x3_reduce = nn.ReLU()
        self.inception_4d_3x3 = nn.Conv2d(
            128, 192, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4d_3x3_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_4d_relu_3x3 = nn.ReLU()
        self.inception_4d_double_3x3_reduce = nn.Conv2d(
            608, 160, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4d_double_3x3_reduce_bn = nn.BatchNorm2d(
            160, momentum=0.1)
        self.inception_4d_relu_double_3x3_reduce = nn.ReLU()
        self.inception_4d_double_3x3_1 = nn.Conv2d(
            160, 192, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4d_double_3x3_1_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_4d_relu_double_3x3_1 = nn.ReLU()
        self.inception_4d_double_3x3_2 = nn.Conv2d(
            192, 192, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4d_double_3x3_2_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_4d_relu_double_3x3_2 = nn.ReLU()
        self.inception_4d_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_4d_pool_proj = nn.Conv2d(
            608, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4d_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4d_relu_pool_proj = nn.ReLU()
        self.inception_4e_3x3_reduce = nn.Conv2d(
            608, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4e_3x3_reduce_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_4e_relu_3x3_reduce = nn.ReLU()
        self.inception_4e_3x3 = nn.Conv2d(
            128, 192, kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4e_3x3_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_4e_relu_3x3 = nn.ReLU()
        self.inception_4e_double_3x3_reduce = nn.Conv2d(
            608, 192, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_4e_double_3x3_reduce_bn = nn.BatchNorm2d(
            192, momentum=0.1)
        self.inception_4e_relu_double_3x3_reduce = nn.ReLU()
        self.inception_4e_double_3x3_1 = nn.Conv2d(
            192, 256, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4e_double_3x3_1_bn = nn.BatchNorm2d(256, momentum=0.1)
        self.inception_4e_relu_double_3x3_1 = nn.ReLU()
        self.inception_4e_double_3x3_2 = nn.Conv2d(
            256, 256, kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_4e_double_3x3_2_bn = nn.BatchNorm2d(256, momentum=0.1)
        self.inception_4e_relu_double_3x3_2 = nn.ReLU()
        self.inception_4e_pool = nn.MaxPool2d(
            kernel_size=(
                3, 3), stride=(
                2, 2), padding=(
                0, 0), ceil_mode=True)
        self.inception_5a_1x1 = nn.Conv2d(
            1056, 352, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5a_1x1_bn = nn.BatchNorm2d(352, momentum=0.1)
        self.inception_5a_relu_1x1 = nn.ReLU()
        self.inception_5a_3x3_reduce = nn.Conv2d(
            1056, 192, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5a_3x3_reduce_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_5a_relu_3x3_reduce = nn.ReLU()
        self.inception_5a_3x3 = nn.Conv2d(
            192, 320, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5a_3x3_bn = nn.BatchNorm2d(320, momentum=0.1)
        self.inception_5a_relu_3x3 = nn.ReLU()
        self.inception_5a_double_3x3_reduce = nn.Conv2d(
            1056, 160, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5a_double_3x3_reduce_bn = nn.BatchNorm2d(
            160, momentum=0.1)
        self.inception_5a_relu_double_3x3_reduce = nn.ReLU()
        self.inception_5a_double_3x3_1 = nn.Conv2d(
            160, 224, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5a_double_3x3_1_bn = nn.BatchNorm2d(224, momentum=0.1)
        self.inception_5a_relu_double_3x3_1 = nn.ReLU()
        self.inception_5a_double_3x3_2 = nn.Conv2d(
            224, 224, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5a_double_3x3_2_bn = nn.BatchNorm2d(224, momentum=0.1)
        self.inception_5a_relu_double_3x3_2 = nn.ReLU()
        self.inception_5a_pool = nn.AvgPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_5a_pool_proj = nn.Conv2d(
            1056, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5a_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_5a_relu_pool_proj = nn.ReLU()
        self.inception_5b_1x1 = nn.Conv2d(
            1024, 352, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5b_1x1_bn = nn.BatchNorm2d(352, momentum=0.1)
        self.inception_5b_relu_1x1 = nn.ReLU()
        self.inception_5b_3x3_reduce = nn.Conv2d(
            1024, 192, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5b_3x3_reduce_bn = nn.BatchNorm2d(192, momentum=0.1)
        self.inception_5b_relu_3x3_reduce = nn.ReLU()
        self.inception_5b_3x3 = nn.Conv2d(
            192, 320, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5b_3x3_bn = nn.BatchNorm2d(320, momentum=0.1)
        self.inception_5b_relu_3x3 = nn.ReLU()
        self.inception_5b_double_3x3_reduce = nn.Conv2d(
            1024, 192, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5b_double_3x3_reduce_bn = nn.BatchNorm2d(
            192, momentum=0.1)
        self.inception_5b_relu_double_3x3_reduce = nn.ReLU()
        self.inception_5b_double_3x3_1 = nn.Conv2d(
            192, 224, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5b_double_3x3_1_bn = nn.BatchNorm2d(224, momentum=0.1)
        self.inception_5b_relu_double_3x3_1 = nn.ReLU()
        self.inception_5b_double_3x3_2 = nn.Conv2d(
            224, 224, kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), dilation=1, groups=1, bias=True)
        self.inception_5b_double_3x3_2_bn = nn.BatchNorm2d(224, momentum=0.1)
        self.inception_5b_relu_double_3x3_2 = nn.ReLU()
        self.inception_5b_pool = nn.MaxPool2d(
            kernel_size=(
                3, 3), stride=(
                1, 1), padding=(
                1, 1), ceil_mode=True)
        self.inception_5b_pool_proj = nn.Conv2d(
            1024, 128, kernel_size=(
                1, 1), stride=(
                1, 1), padding=(
                0, 0), dilation=1, groups=1, bias=True)
        self.inception_5b_pool_proj_bn = nn.BatchNorm2d(128, momentum=0.1)
        self.inception_5b_relu_pool_proj = nn.ReLU()
        # self.global_pool = nn.AvgPool2d(kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), ceil_mode=True)
        self.dropout = nn.Dropout(0.699999988079)
        self.flatten = nn.Flatten()
        self.fc_action = nn.Linear(1024, 200)

        if pretrained_model:
            state_dict = torch.load(get_bn_inception_pretrained_model())
            self.load_state_dict(state_dict)

    def forward(self, data):
        conv1_7x7_s2 = self.conv1_7x7_s2(data)
        conv1_7x7_s2_bn = self.conv1_7x7_s2_bn(conv1_7x7_s2)
        conv1_7x7_s2_bn = self.conv1_relu_7x7(conv1_7x7_s2_bn)
        pool1_3x3_s2 = self.pool1_3x3_s2(conv1_7x7_s2_bn)
        conv2_3x3_reduce = self.conv2_3x3_reduce(pool1_3x3_s2)
        conv2_3x3_reduce_bn = self.conv2_3x3_reduce_bn(conv2_3x3_reduce)
        conv2_3x3_reduce_bn = self.conv2_relu_3x3_reduce(conv2_3x3_reduce_bn)
        conv2_3x3 = self.conv2_3x3(conv2_3x3_reduce_bn)
        conv2_3x3_bn = self.conv2_3x3_bn(conv2_3x3)
        conv2_3x3_bn = self.conv2_relu_3x3(conv2_3x3_bn)
        pool2_3x3_s2 = self.pool2_3x3_s2(conv2_3x3_bn)
        pool2_3x3_s2_pool2_3x3_s2_0_split_0 = pool2_3x3_s2
        pool2_3x3_s2_pool2_3x3_s2_0_split_1 = pool2_3x3_s2
        pool2_3x3_s2_pool2_3x3_s2_0_split_2 = pool2_3x3_s2
        pool2_3x3_s2_pool2_3x3_s2_0_split_3 = pool2_3x3_s2
        inception_3a_1x1 = self.inception_3a_1x1(
            pool2_3x3_s2_pool2_3x3_s2_0_split_0)
        inception_3a_1x1_bn = self.inception_3a_1x1_bn(inception_3a_1x1)
        inception_3a_1x1_bn = self.inception_3a_relu_1x1(inception_3a_1x1_bn)
        inception_3a_3x3_reduce = self.inception_3a_3x3_reduce(
            pool2_3x3_s2_pool2_3x3_s2_0_split_1)
        inception_3a_3x3_reduce_bn = self.inception_3a_3x3_reduce_bn(
            inception_3a_3x3_reduce)
        inception_3a_3x3_reduce_bn = self.inception_3a_relu_3x3_reduce(
            inception_3a_3x3_reduce_bn)
        inception_3a_3x3 = self.inception_3a_3x3(inception_3a_3x3_reduce_bn)
        inception_3a_3x3_bn = self.inception_3a_3x3_bn(inception_3a_3x3)
        inception_3a_3x3_bn = self.inception_3a_relu_3x3(inception_3a_3x3_bn)
        inception_3a_double_3x3_reduce = self.inception_3a_double_3x3_reduce(
            pool2_3x3_s2_pool2_3x3_s2_0_split_2)
        inception_3a_double_3x3_reduce_bn = self.inception_3a_double_3x3_reduce_bn(
            inception_3a_double_3x3_reduce)
        inception_3a_double_3x3_reduce_bn = self.inception_3a_relu_double_3x3_reduce(
            inception_3a_double_3x3_reduce_bn)
        inception_3a_double_3x3_1 = self.inception_3a_double_3x3_1(
            inception_3a_double_3x3_reduce_bn)
        inception_3a_double_3x3_1_bn = self.inception_3a_double_3x3_1_bn(
            inception_3a_double_3x3_1)
        inception_3a_double_3x3_1_bn = self.inception_3a_relu_double_3x3_1(
            inception_3a_double_3x3_1_bn)
        inception_3a_double_3x3_2 = self.inception_3a_double_3x3_2(
            inception_3a_double_3x3_1_bn)
        inception_3a_double_3x3_2_bn = self.inception_3a_double_3x3_2_bn(
            inception_3a_double_3x3_2)
        inception_3a_double_3x3_2_bn = self.inception_3a_relu_double_3x3_2(
            inception_3a_double_3x3_2_bn)
        inception_3a_pool = self.inception_3a_pool(
            pool2_3x3_s2_pool2_3x3_s2_0_split_3)
        inception_3a_pool_proj = self.inception_3a_pool_proj(inception_3a_pool)
        inception_3a_pool_proj_bn = self.inception_3a_pool_proj_bn(
            inception_3a_pool_proj)
        inception_3a_pool_proj_bn = self.inception_3a_relu_pool_proj(
            inception_3a_pool_proj_bn)
        inception_3a_output = torch.cat(
            (inception_3a_1x1_bn,
             inception_3a_3x3_bn,
             inception_3a_double_3x3_2_bn,
             inception_3a_pool_proj_bn),
            dim=1)
        inception_3a_output_inception_3a_output_0_split_0 = inception_3a_output
        inception_3a_output_inception_3a_output_0_split_1 = inception_3a_output
        inception_3a_output_inception_3a_output_0_split_2 = inception_3a_output
        inception_3a_output_inception_3a_output_0_split_3 = inception_3a_output
        inception_3b_1x1 = self.inception_3b_1x1(
            inception_3a_output_inception_3a_output_0_split_0)
        inception_3b_1x1_bn = self.inception_3b_1x1_bn(inception_3b_1x1)
        inception_3b_1x1_bn = self.inception_3b_relu_1x1(inception_3b_1x1_bn)
        inception_3b_3x3_reduce = self.inception_3b_3x3_reduce(
            inception_3a_output_inception_3a_output_0_split_1)
        inception_3b_3x3_reduce_bn = self.inception_3b_3x3_reduce_bn(
            inception_3b_3x3_reduce)
        inception_3b_3x3_reduce_bn = self.inception_3b_relu_3x3_reduce(
            inception_3b_3x3_reduce_bn)
        inception_3b_3x3 = self.inception_3b_3x3(inception_3b_3x3_reduce_bn)
        inception_3b_3x3_bn = self.inception_3b_3x3_bn(inception_3b_3x3)
        inception_3b_3x3_bn = self.inception_3b_relu_3x3(inception_3b_3x3_bn)
        inception_3b_double_3x3_reduce = self.inception_3b_double_3x3_reduce(
            inception_3a_output_inception_3a_output_0_split_2)
        inception_3b_double_3x3_reduce_bn = self.inception_3b_double_3x3_reduce_bn(
            inception_3b_double_3x3_reduce)
        inception_3b_double_3x3_reduce_bn = self.inception_3b_relu_double_3x3_reduce(
            inception_3b_double_3x3_reduce_bn)
        inception_3b_double_3x3_1 = self.inception_3b_double_3x3_1(
            inception_3b_double_3x3_reduce_bn)
        inception_3b_double_3x3_1_bn = self.inception_3b_double_3x3_1_bn(
            inception_3b_double_3x3_1)
        inception_3b_double_3x3_1_bn = self.inception_3b_relu_double_3x3_1(
            inception_3b_double_3x3_1_bn)
        inception_3b_double_3x3_2 = self.inception_3b_double_3x3_2(
            inception_3b_double_3x3_1_bn)
        inception_3b_double_3x3_2_bn = self.inception_3b_double_3x3_2_bn(
            inception_3b_double_3x3_2)
        inception_3b_double_3x3_2_bn = self.inception_3b_relu_double_3x3_2(
            inception_3b_double_3x3_2_bn)
        inception_3b_pool = self.inception_3b_pool(
            inception_3a_output_inception_3a_output_0_split_3)
        inception_3b_pool_proj = self.inception_3b_pool_proj(inception_3b_pool)
        inception_3b_pool_proj_bn = self.inception_3b_pool_proj_bn(
            inception_3b_pool_proj)
        inception_3b_pool_proj_bn = self.inception_3b_relu_pool_proj(
            inception_3b_pool_proj_bn)
        inception_3b_output = torch.cat(
            (inception_3b_1x1_bn,
             inception_3b_3x3_bn,
             inception_3b_double_3x3_2_bn,
             inception_3b_pool_proj_bn),
            dim=1)
        inception_3b_output_inception_3b_output_0_split_0 = inception_3b_output
        inception_3b_output_inception_3b_output_0_split_1 = inception_3b_output
        inception_3b_output_inception_3b_output_0_split_2 = inception_3b_output
        inception_3c_3x3_reduce = self.inception_3c_3x3_reduce(
            inception_3b_output_inception_3b_output_0_split_0)
        inception_3c_3x3_reduce_bn = self.inception_3c_3x3_reduce_bn(
            inception_3c_3x3_reduce)
        inception_3c_3x3_reduce_bn = self.inception_3c_relu_3x3_reduce(
            inception_3c_3x3_reduce_bn)
        inception_3c_3x3 = self.inception_3c_3x3(inception_3c_3x3_reduce_bn)
        inception_3c_3x3_bn = self.inception_3c_3x3_bn(inception_3c_3x3)
        inception_3c_3x3_bn = self.inception_3c_relu_3x3(inception_3c_3x3_bn)
        inception_3c_double_3x3_reduce = self.inception_3c_double_3x3_reduce(
            inception_3b_output_inception_3b_output_0_split_1)
        inception_3c_double_3x3_reduce_bn = self.inception_3c_double_3x3_reduce_bn(
            inception_3c_double_3x3_reduce)
        inception_3c_double_3x3_reduce_bn = self.inception_3c_relu_double_3x3_reduce(
            inception_3c_double_3x3_reduce_bn)
        inception_3c_double_3x3_1 = self.inception_3c_double_3x3_1(
            inception_3c_double_3x3_reduce_bn)
        inception_3c_double_3x3_1_bn = self.inception_3c_double_3x3_1_bn(
            inception_3c_double_3x3_1)
        inception_3c_double_3x3_1_bn = self.inception_3c_relu_double_3x3_1(
            inception_3c_double_3x3_1_bn)
        inception_3c_double_3x3_2 = self.inception_3c_double_3x3_2(
            inception_3c_double_3x3_1_bn)
        inception_3c_double_3x3_2_bn = self.inception_3c_double_3x3_2_bn(
            inception_3c_double_3x3_2)
        inception_3c_double_3x3_2_bn = self.inception_3c_relu_double_3x3_2(
            inception_3c_double_3x3_2_bn)
        inception_3c_pool = self.inception_3c_pool(
            inception_3b_output_inception_3b_output_0_split_2)
        inception_3c_output = torch.cat(
            (inception_3c_3x3_bn,
             inception_3c_double_3x3_2_bn,
             inception_3c_pool),
            dim=1)
        inception_3c_output_inception_3c_output_0_split_0 = inception_3c_output
        inception_3c_output_inception_3c_output_0_split_1 = inception_3c_output
        inception_3c_output_inception_3c_output_0_split_2 = inception_3c_output
        inception_3c_output_inception_3c_output_0_split_3 = inception_3c_output
        inception_4a_1x1 = self.inception_4a_1x1(
            inception_3c_output_inception_3c_output_0_split_0)
        inception_4a_1x1_bn = self.inception_4a_1x1_bn(inception_4a_1x1)
        inception_4a_1x1_bn = self.inception_4a_relu_1x1(inception_4a_1x1_bn)
        inception_4a_3x3_reduce = self.inception_4a_3x3_reduce(
            inception_3c_output_inception_3c_output_0_split_1)
        inception_4a_3x3_reduce_bn = self.inception_4a_3x3_reduce_bn(
            inception_4a_3x3_reduce)
        inception_4a_3x3_reduce_bn = self.inception_4a_relu_3x3_reduce(
            inception_4a_3x3_reduce_bn)
        inception_4a_3x3 = self.inception_4a_3x3(inception_4a_3x3_reduce_bn)
        inception_4a_3x3_bn = self.inception_4a_3x3_bn(inception_4a_3x3)
        inception_4a_3x3_bn = self.inception_4a_relu_3x3(inception_4a_3x3_bn)
        inception_4a_double_3x3_reduce = self.inception_4a_double_3x3_reduce(
            inception_3c_output_inception_3c_output_0_split_2)
        inception_4a_double_3x3_reduce_bn = self.inception_4a_double_3x3_reduce_bn(
            inception_4a_double_3x3_reduce)
        inception_4a_double_3x3_reduce_bn = self.inception_4a_relu_double_3x3_reduce(
            inception_4a_double_3x3_reduce_bn)
        inception_4a_double_3x3_1 = self.inception_4a_double_3x3_1(
            inception_4a_double_3x3_reduce_bn)
        inception_4a_double_3x3_1_bn = self.inception_4a_double_3x3_1_bn(
            inception_4a_double_3x3_1)
        inception_4a_double_3x3_1_bn = self.inception_4a_relu_double_3x3_1(
            inception_4a_double_3x3_1_bn)
        inception_4a_double_3x3_2 = self.inception_4a_double_3x3_2(
            inception_4a_double_3x3_1_bn)
        inception_4a_double_3x3_2_bn = self.inception_4a_double_3x3_2_bn(
            inception_4a_double_3x3_2)
        inception_4a_double_3x3_2_bn = self.inception_4a_relu_double_3x3_2(
            inception_4a_double_3x3_2_bn)
        inception_4a_pool = self.inception_4a_pool(
            inception_3c_output_inception_3c_output_0_split_3)
        inception_4a_pool_proj = self.inception_4a_pool_proj(inception_4a_pool)
        inception_4a_pool_proj_bn = self.inception_4a_pool_proj_bn(
            inception_4a_pool_proj)
        inception_4a_pool_proj_bn = self.inception_4a_relu_pool_proj(
            inception_4a_pool_proj_bn)
        inception_4a_output = torch.cat(
            (inception_4a_1x1_bn,
             inception_4a_3x3_bn,
             inception_4a_double_3x3_2_bn,
             inception_4a_pool_proj_bn),
            dim=1)
        inception_4a_output_inception_4a_output_0_split_0 = inception_4a_output
        inception_4a_output_inception_4a_output_0_split_1 = inception_4a_output
        inception_4a_output_inception_4a_output_0_split_2 = inception_4a_output
        inception_4a_output_inception_4a_output_0_split_3 = inception_4a_output
        inception_4b_1x1 = self.inception_4b_1x1(
            inception_4a_output_inception_4a_output_0_split_0)
        inception_4b_1x1_bn = self.inception_4b_1x1_bn(inception_4b_1x1)
        inception_4b_1x1_bn = self.inception_4b_relu_1x1(inception_4b_1x1_bn)
        inception_4b_3x3_reduce = self.inception_4b_3x3_reduce(
            inception_4a_output_inception_4a_output_0_split_1)
        inception_4b_3x3_reduce_bn = self.inception_4b_3x3_reduce_bn(
            inception_4b_3x3_reduce)
        inception_4b_3x3_reduce_bn = self.inception_4b_relu_3x3_reduce(
            inception_4b_3x3_reduce_bn)
        inception_4b_3x3 = self.inception_4b_3x3(inception_4b_3x3_reduce_bn)
        inception_4b_3x3_bn = self.inception_4b_3x3_bn(inception_4b_3x3)
        inception_4b_3x3_bn = self.inception_4b_relu_3x3(inception_4b_3x3_bn)
        inception_4b_double_3x3_reduce = self.inception_4b_double_3x3_reduce(
            inception_4a_output_inception_4a_output_0_split_2)
        inception_4b_double_3x3_reduce_bn = self.inception_4b_double_3x3_reduce_bn(
            inception_4b_double_3x3_reduce)
        inception_4b_double_3x3_reduce_bn = self.inception_4b_relu_double_3x3_reduce(
            inception_4b_double_3x3_reduce_bn)
        inception_4b_double_3x3_1 = self.inception_4b_double_3x3_1(
            inception_4b_double_3x3_reduce_bn)
        inception_4b_double_3x3_1_bn = self.inception_4b_double_3x3_1_bn(
            inception_4b_double_3x3_1)
        inception_4b_double_3x3_1_bn = self.inception_4b_relu_double_3x3_1(
            inception_4b_double_3x3_1_bn)
        inception_4b_double_3x3_2 = self.inception_4b_double_3x3_2(
            inception_4b_double_3x3_1_bn)
        inception_4b_double_3x3_2_bn = self.inception_4b_double_3x3_2_bn(
            inception_4b_double_3x3_2)
        inception_4b_double_3x3_2_bn = self.inception_4b_relu_double_3x3_2(
            inception_4b_double_3x3_2_bn)
        inception_4b_pool = self.inception_4b_pool(
            inception_4a_output_inception_4a_output_0_split_3)
        inception_4b_pool_proj = self.inception_4b_pool_proj(inception_4b_pool)
        inception_4b_pool_proj_bn = self.inception_4b_pool_proj_bn(
            inception_4b_pool_proj)
        inception_4b_pool_proj_bn = self.inception_4b_relu_pool_proj(
            inception_4b_pool_proj_bn)
        inception_4b_output = torch.cat(
            (inception_4b_1x1_bn,
             inception_4b_3x3_bn,
             inception_4b_double_3x3_2_bn,
             inception_4b_pool_proj_bn),
            dim=1)
        inception_4b_output_inception_4b_output_0_split_0 = inception_4b_output
        inception_4b_output_inception_4b_output_0_split_1 = inception_4b_output
        inception_4b_output_inception_4b_output_0_split_2 = inception_4b_output
        inception_4b_output_inception_4b_output_0_split_3 = inception_4b_output
        inception_4c_1x1 = self.inception_4c_1x1(
            inception_4b_output_inception_4b_output_0_split_0)
        inception_4c_1x1_bn = self.inception_4c_1x1_bn(inception_4c_1x1)
        inception_4c_1x1_bn = self.inception_4c_relu_1x1(inception_4c_1x1_bn)
        inception_4c_3x3_reduce = self.inception_4c_3x3_reduce(
            inception_4b_output_inception_4b_output_0_split_1)
        inception_4c_3x3_reduce_bn = self.inception_4c_3x3_reduce_bn(
            inception_4c_3x3_reduce)
        inception_4c_3x3_reduce_bn = self.inception_4c_relu_3x3_reduce(
            inception_4c_3x3_reduce_bn)
        inception_4c_3x3 = self.inception_4c_3x3(inception_4c_3x3_reduce_bn)
        inception_4c_3x3_bn = self.inception_4c_3x3_bn(inception_4c_3x3)
        inception_4c_3x3_bn = self.inception_4c_relu_3x3(inception_4c_3x3_bn)
        inception_4c_double_3x3_reduce = self.inception_4c_double_3x3_reduce(
            inception_4b_output_inception_4b_output_0_split_2)
        inception_4c_double_3x3_reduce_bn = self.inception_4c_double_3x3_reduce_bn(
            inception_4c_double_3x3_reduce)
        inception_4c_double_3x3_reduce_bn = self.inception_4c_relu_double_3x3_reduce(
            inception_4c_double_3x3_reduce_bn)
        inception_4c_double_3x3_1 = self.inception_4c_double_3x3_1(
            inception_4c_double_3x3_reduce_bn)
        inception_4c_double_3x3_1_bn = self.inception_4c_double_3x3_1_bn(
            inception_4c_double_3x3_1)
        inception_4c_double_3x3_1_bn = self.inception_4c_relu_double_3x3_1(
            inception_4c_double_3x3_1_bn)
        inception_4c_double_3x3_2 = self.inception_4c_double_3x3_2(
            inception_4c_double_3x3_1_bn)
        inception_4c_double_3x3_2_bn = self.inception_4c_double_3x3_2_bn(
            inception_4c_double_3x3_2)
        inception_4c_double_3x3_2_bn = self.inception_4c_relu_double_3x3_2(
            inception_4c_double_3x3_2_bn)
        inception_4c_pool = self.inception_4c_pool(
            inception_4b_output_inception_4b_output_0_split_3)
        inception_4c_pool_proj = self.inception_4c_pool_proj(inception_4c_pool)
        inception_4c_pool_proj_bn = self.inception_4c_pool_proj_bn(
            inception_4c_pool_proj)
        inception_4c_pool_proj_bn = self.inception_4c_relu_pool_proj(
            inception_4c_pool_proj_bn)
        inception_4c_output = torch.cat(
            (inception_4c_1x1_bn,
             inception_4c_3x3_bn,
             inception_4c_double_3x3_2_bn,
             inception_4c_pool_proj_bn),
            dim=1)
        inception_4c_output_inception_4c_output_0_split_0 = inception_4c_output
        inception_4c_output_inception_4c_output_0_split_1 = inception_4c_output
        inception_4c_output_inception_4c_output_0_split_2 = inception_4c_output
        inception_4c_output_inception_4c_output_0_split_3 = inception_4c_output
        inception_4d_1x1 = self.inception_4d_1x1(
            inception_4c_output_inception_4c_output_0_split_0)
        inception_4d_1x1_bn = self.inception_4d_1x1_bn(inception_4d_1x1)
        inception_4d_1x1_bn = self.inception_4d_relu_1x1(inception_4d_1x1_bn)
        inception_4d_3x3_reduce = self.inception_4d_3x3_reduce(
            inception_4c_output_inception_4c_output_0_split_1)
        inception_4d_3x3_reduce_bn = self.inception_4d_3x3_reduce_bn(
            inception_4d_3x3_reduce)
        inception_4d_3x3_reduce_bn = self.inception_4d_relu_3x3_reduce(
            inception_4d_3x3_reduce_bn)
        inception_4d_3x3 = self.inception_4d_3x3(inception_4d_3x3_reduce_bn)
        inception_4d_3x3_bn = self.inception_4d_3x3_bn(inception_4d_3x3)
        inception_4d_3x3_bn = self.inception_4d_relu_3x3(inception_4d_3x3_bn)
        inception_4d_double_3x3_reduce = self.inception_4d_double_3x3_reduce(
            inception_4c_output_inception_4c_output_0_split_2)
        inception_4d_double_3x3_reduce_bn = self.inception_4d_double_3x3_reduce_bn(
            inception_4d_double_3x3_reduce)
        inception_4d_double_3x3_reduce_bn = self.inception_4d_relu_double_3x3_reduce(
            inception_4d_double_3x3_reduce_bn)
        inception_4d_double_3x3_1 = self.inception_4d_double_3x3_1(
            inception_4d_double_3x3_reduce_bn)
        inception_4d_double_3x3_1_bn = self.inception_4d_double_3x3_1_bn(
            inception_4d_double_3x3_1)
        inception_4d_double_3x3_1_bn = self.inception_4d_relu_double_3x3_1(
            inception_4d_double_3x3_1_bn)
        inception_4d_double_3x3_2 = self.inception_4d_double_3x3_2(
            inception_4d_double_3x3_1_bn)
        inception_4d_double_3x3_2_bn = self.inception_4d_double_3x3_2_bn(
            inception_4d_double_3x3_2)
        inception_4d_double_3x3_2_bn = self.inception_4d_relu_double_3x3_2(
            inception_4d_double_3x3_2_bn)
        inception_4d_pool = self.inception_4d_pool(
            inception_4c_output_inception_4c_output_0_split_3)
        inception_4d_pool_proj = self.inception_4d_pool_proj(inception_4d_pool)
        inception_4d_pool_proj_bn = self.inception_4d_pool_proj_bn(
            inception_4d_pool_proj)
        inception_4d_pool_proj_bn = self.inception_4d_relu_pool_proj(
            inception_4d_pool_proj_bn)
        inception_4d_output = torch.cat(
            (inception_4d_1x1_bn,
             inception_4d_3x3_bn,
             inception_4d_double_3x3_2_bn,
             inception_4d_pool_proj_bn),
            dim=1)
        inception_4d_output_inception_4d_output_0_split_0 = inception_4d_output
        inception_4d_output_inception_4d_output_0_split_1 = inception_4d_output
        inception_4d_output_inception_4d_output_0_split_2 = inception_4d_output
        inception_4e_3x3_reduce = self.inception_4e_3x3_reduce(
            inception_4d_output_inception_4d_output_0_split_0)
        inception_4e_3x3_reduce_bn = self.inception_4e_3x3_reduce_bn(
            inception_4e_3x3_reduce)
        inception_4e_3x3_reduce_bn = self.inception_4e_relu_3x3_reduce(
            inception_4e_3x3_reduce_bn)
        inception_4e_3x3 = self.inception_4e_3x3(inception_4e_3x3_reduce_bn)
        inception_4e_3x3_bn = self.inception_4e_3x3_bn(inception_4e_3x3)
        inception_4e_3x3_bn = self.inception_4e_relu_3x3(inception_4e_3x3_bn)
        inception_4e_double_3x3_reduce = self.inception_4e_double_3x3_reduce(
            inception_4d_output_inception_4d_output_0_split_1)
        inception_4e_double_3x3_reduce_bn = self.inception_4e_double_3x3_reduce_bn(
            inception_4e_double_3x3_reduce)
        inception_4e_double_3x3_reduce_bn = self.inception_4e_relu_double_3x3_reduce(
            inception_4e_double_3x3_reduce_bn)
        inception_4e_double_3x3_1 = self.inception_4e_double_3x3_1(
            inception_4e_double_3x3_reduce_bn)
        inception_4e_double_3x3_1_bn = self.inception_4e_double_3x3_1_bn(
            inception_4e_double_3x3_1)
        inception_4e_double_3x3_1_bn = self.inception_4e_relu_double_3x3_1(
            inception_4e_double_3x3_1_bn)
        inception_4e_double_3x3_2 = self.inception_4e_double_3x3_2(
            inception_4e_double_3x3_1_bn)
        inception_4e_double_3x3_2_bn = self.inception_4e_double_3x3_2_bn(
            inception_4e_double_3x3_2)
        inception_4e_double_3x3_2_bn = self.inception_4e_relu_double_3x3_2(
            inception_4e_double_3x3_2_bn)
        inception_4e_pool = self.inception_4e_pool(
            inception_4d_output_inception_4d_output_0_split_2)
        inception_4e_output = torch.cat(
            (inception_4e_3x3_bn,
             inception_4e_double_3x3_2_bn,
             inception_4e_pool),
            dim=1)
        inception_4e_output_inception_4e_output_0_split_0 = inception_4e_output
        inception_4e_output_inception_4e_output_0_split_1 = inception_4e_output
        inception_4e_output_inception_4e_output_0_split_2 = inception_4e_output
        inception_4e_output_inception_4e_output_0_split_3 = inception_4e_output
        inception_5a_1x1 = self.inception_5a_1x1(
            inception_4e_output_inception_4e_output_0_split_0)
        inception_5a_1x1_bn = self.inception_5a_1x1_bn(inception_5a_1x1)
        inception_5a_1x1_bn = self.inception_5a_relu_1x1(inception_5a_1x1_bn)
        inception_5a_3x3_reduce = self.inception_5a_3x3_reduce(
            inception_4e_output_inception_4e_output_0_split_1)
        inception_5a_3x3_reduce_bn = self.inception_5a_3x3_reduce_bn(
            inception_5a_3x3_reduce)
        inception_5a_3x3_reduce_bn = self.inception_5a_relu_3x3_reduce(
            inception_5a_3x3_reduce_bn)
        inception_5a_3x3 = self.inception_5a_3x3(inception_5a_3x3_reduce_bn)
        inception_5a_3x3_bn = self.inception_5a_3x3_bn(inception_5a_3x3)
        inception_5a_3x3_bn = self.inception_5a_relu_3x3(inception_5a_3x3_bn)
        inception_5a_double_3x3_reduce = self.inception_5a_double_3x3_reduce(
            inception_4e_output_inception_4e_output_0_split_2)
        inception_5a_double_3x3_reduce_bn = self.inception_5a_double_3x3_reduce_bn(
            inception_5a_double_3x3_reduce)
        inception_5a_double_3x3_reduce_bn = self.inception_5a_relu_double_3x3_reduce(
            inception_5a_double_3x3_reduce_bn)
        inception_5a_double_3x3_1 = self.inception_5a_double_3x3_1(
            inception_5a_double_3x3_reduce_bn)
        inception_5a_double_3x3_1_bn = self.inception_5a_double_3x3_1_bn(
            inception_5a_double_3x3_1)
        inception_5a_double_3x3_1_bn = self.inception_5a_relu_double_3x3_1(
            inception_5a_double_3x3_1_bn)
        inception_5a_double_3x3_2 = self.inception_5a_double_3x3_2(
            inception_5a_double_3x3_1_bn)
        inception_5a_double_3x3_2_bn = self.inception_5a_double_3x3_2_bn(
            inception_5a_double_3x3_2)
        inception_5a_double_3x3_2_bn = self.inception_5a_relu_double_3x3_2(
            inception_5a_double_3x3_2_bn)
        inception_5a_pool = self.inception_5a_pool(
            inception_4e_output_inception_4e_output_0_split_3)
        inception_5a_pool_proj = self.inception_5a_pool_proj(inception_5a_pool)
        inception_5a_pool_proj_bn = self.inception_5a_pool_proj_bn(
            inception_5a_pool_proj)
        inception_5a_pool_proj_bn = self.inception_5a_relu_pool_proj(
            inception_5a_pool_proj_bn)
        inception_5a_output = torch.cat(
            (inception_5a_1x1_bn,
             inception_5a_3x3_bn,
             inception_5a_double_3x3_2_bn,
             inception_5a_pool_proj_bn),
            dim=1)
        inception_5a_output_inception_5a_output_0_split_0 = inception_5a_output
        inception_5a_output_inception_5a_output_0_split_1 = inception_5a_output
        inception_5a_output_inception_5a_output_0_split_2 = inception_5a_output
        inception_5a_output_inception_5a_output_0_split_3 = inception_5a_output
        inception_5b_1x1 = self.inception_5b_1x1(
            inception_5a_output_inception_5a_output_0_split_0)
        inception_5b_1x1_bn = self.inception_5b_1x1_bn(inception_5b_1x1)
        inception_5b_1x1_bn = self.inception_5b_relu_1x1(inception_5b_1x1_bn)
        inception_5b_3x3_reduce = self.inception_5b_3x3_reduce(
            inception_5a_output_inception_5a_output_0_split_1)
        inception_5b_3x3_reduce_bn = self.inception_5b_3x3_reduce_bn(
            inception_5b_3x3_reduce)
        inception_5b_3x3_reduce_bn = self.inception_5b_relu_3x3_reduce(
            inception_5b_3x3_reduce_bn)
        inception_5b_3x3 = self.inception_5b_3x3(inception_5b_3x3_reduce_bn)
        inception_5b_3x3_bn = self.inception_5b_3x3_bn(inception_5b_3x3)
        inception_5b_3x3_bn = self.inception_5b_relu_3x3(inception_5b_3x3_bn)
        inception_5b_double_3x3_reduce = self.inception_5b_double_3x3_reduce(
            inception_5a_output_inception_5a_output_0_split_2)
        inception_5b_double_3x3_reduce_bn = self.inception_5b_double_3x3_reduce_bn(
            inception_5b_double_3x3_reduce)
        inception_5b_double_3x3_reduce_bn = self.inception_5b_relu_double_3x3_reduce(
            inception_5b_double_3x3_reduce_bn)
        inception_5b_double_3x3_1 = self.inception_5b_double_3x3_1(
            inception_5b_double_3x3_reduce_bn)
        inception_5b_double_3x3_1_bn = self.inception_5b_double_3x3_1_bn(
            inception_5b_double_3x3_1)
        inception_5b_double_3x3_1_bn = self.inception_5b_relu_double_3x3_1(
            inception_5b_double_3x3_1_bn)
        inception_5b_double_3x3_2 = self.inception_5b_double_3x3_2(
            inception_5b_double_3x3_1_bn)
        inception_5b_double_3x3_2_bn = self.inception_5b_double_3x3_2_bn(
            inception_5b_double_3x3_2)
        inception_5b_double_3x3_2_bn = self.inception_5b_relu_double_3x3_2(
            inception_5b_double_3x3_2_bn)
        inception_5b_pool = self.inception_5b_pool(
            inception_5a_output_inception_5a_output_0_split_3)
        inception_5b_pool_proj = self.inception_5b_pool_proj(inception_5b_pool)
        inception_5b_pool_proj_bn = self.inception_5b_pool_proj_bn(
            inception_5b_pool_proj)
        inception_5b_pool_proj_bn = self.inception_5b_relu_pool_proj(
            inception_5b_pool_proj_bn)
        inception_5b_output = torch.cat(
            (inception_5b_1x1_bn,
             inception_5b_3x3_bn,
             inception_5b_double_3x3_2_bn,
             inception_5b_pool_proj_bn),
            dim=1)
        global_pool = F.avg_pool2d(
            inception_5b_output,
            kernel_size=(
                inception_5b_output.shape[2],
                inception_5b_output.shape[3]),
            padding=0)
        global_pool = self.dropout(global_pool)
        global_pool = self.flatten(global_pool)
        return global_pool
        # fc_action = self.fc_action(global_pool)
        # return fc_action



#!/usr/bin/env python2.7

import argparse
import os
import sys
import tempfile

import caffe
import numpy as np
import torch
import torch.nn as nn
from caffe.proto import caffe_pb2

parser = argparse.ArgumentParser()
parser.add_argument('caffe_model', help='input model in hdf5 or caffemodel format')
parser.add_argument('prototxt_template',help='prototxt template')
parser.add_argument('flownet2_pytorch', help='path to flownet2-pytorch')

args = parser.parse_args()

args.rgb_max = 255
args.fp16 = False
args.grads = {}

# load models
sys.path.append(args.flownet2_pytorch)

import models
from utils.param_utils import *

width = 256
height = 256
keys = {'TARGET_WIDTH': width, 
        'TARGET_HEIGHT': height,
        'ADAPTED_WIDTH':width,
        'ADAPTED_HEIGHT':height,
        'SCALE_WIDTH':1.,
        'SCALE_HEIGHT':1.,}

template = '\n'.join(np.loadtxt(args.prototxt_template, dtype=str, delimiter='\n'))
for k in keys:
    template = template.replace('$%s$'%(k),str(keys[k]))

prototxt = tempfile.NamedTemporaryFile(mode='w', delete=True)
prototxt.write(template)
prototxt.flush()

net = caffe.Net(prototxt.name, args.caffe_model, caffe.TEST)

weights = {}
biases = {}

for k, v in list(net.params.items()):
    weights[k] = np.array(v[0].data).reshape(v[0].data.shape)
    biases[k] = np.array(v[1].data).reshape(v[1].data.shape)
    print((k, weights[k].shape, biases[k].shape))

if 'FlowNet2/' in args.caffe_model:
    model = models.FlowNet2(args)

    parse_flownetc(model.flownetc.modules(), weights, biases)
    parse_flownets(model.flownets_1.modules(), weights, biases, param_prefix='net2_')
    parse_flownets(model.flownets_2.modules(), weights, biases, param_prefix='net3_')
    parse_flownetsd(model.flownets_d.modules(), weights, biases, param_prefix='netsd_')
    parse_flownetfusion(model.flownetfusion.modules(), weights, biases, param_prefix='fuse_')

    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2_checkpoint.pth.tar'))

elif 'FlowNet2-C/' in args.caffe_model:
    model = models.FlowNet2C(args)

    parse_flownetc(model.modules(), weights, biases)
    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-C_checkpoint.pth.tar'))

elif 'FlowNet2-CS/' in args.caffe_model:
    model = models.FlowNet2CS(args)

    parse_flownetc(model.flownetc.modules(), weights, biases)
    parse_flownets(model.flownets_1.modules(), weights, biases, param_prefix='net2_')

    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-CS_checkpoint.pth.tar'))

elif 'FlowNet2-CSS/' in args.caffe_model:
    model = models.FlowNet2CSS(args)

    parse_flownetc(model.flownetc.modules(), weights, biases)
    parse_flownets(model.flownets_1.modules(), weights, biases, param_prefix='net2_')
    parse_flownets(model.flownets_2.modules(), weights, biases, param_prefix='net3_')

    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-CSS_checkpoint.pth.tar'))

elif 'FlowNet2-CSS-ft-sd/' in args.caffe_model:
    model = models.FlowNet2CSS(args)

    parse_flownetc(model.flownetc.modules(), weights, biases)
    parse_flownets(model.flownets_1.modules(), weights, biases, param_prefix='net2_')
    parse_flownets(model.flownets_2.modules(), weights, biases, param_prefix='net3_')

    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-CSS-ft-sd_checkpoint.pth.tar'))

elif 'FlowNet2-S/' in args.caffe_model:
    model = models.FlowNet2S(args)

    parse_flownetsonly(model.modules(), weights, biases, param_prefix='')
    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-S_checkpoint.pth.tar'))

elif 'FlowNet2-SD/' in args.caffe_model:
    model = models.FlowNet2SD(args)

    parse_flownetsd(model.modules(), weights, biases, param_prefix='')

    state = {'epoch': 0,
             'state_dict': model.state_dict(),
             'best_EPE': 1e10}
    torch.save(state, os.path.join(args.flownet2_pytorch, 'FlowNet2-SD_checkpoint.pth.tar'))

else:
    print(('model type cound not be determined from input caffe model %s'%(args.caffe_model)))
    quit()
print(("done converting ", args.caffe_model))


import os.path as osp

import gdown
from eos import makedirs

download_dir = osp.expanduser('~/.anet_tools')


def get_resnet200_pretrained_model():
    makedirs(download_dir)
    target_path = osp.join(download_dir, 'resnet200_anet_2016.pt')
    if not osp.exists(target_path):
        gdown.cached_download(
            'https://drive.google.com/uc?id=1Xin1jzNNle_R-O-E9nVQDgv086eoj0qs',
            path=target_path,
            md5='010cd5bbccddc90c9a6e3fbed258eeb4')
    return target_path


def get_bn_inception_pretrained_model():
    makedirs(download_dir)
    target_path = osp.join(
        download_dir, 'bn_inception_anet_2016_temporal.v5.pt')
    if not osp.exists(target_path):
        gdown.cached_download(
            'https://drive.google.com/uc?id=1NJ9u10jrqFQwe3AJoKLm7oY-Rf0leKoV',
            path=target_path,
            md5='c4c2bf46584b651e19a6e41df67f4a27')
    return target_path



import torch
from torch.autograd import Function
from torch.nn.modules.module import Module

import correlation_cuda


class CorrelationFunction(Function):

    @staticmethod
    def forward(ctx, input1, input2, pad_size=3, kernel_size=3, max_displacement=20, stride1=1, stride2=2, corr_multiply=1):
        ctx.save_for_backward(input1, input2)

        ctx.pad_size = pad_size
        ctx.kernel_size = kernel_size
        ctx.max_displacement = max_displacement
        ctx.stride1 = stride1
        ctx.stride2 = stride2
        ctx.corr_multiply = corr_multiply

        with torch.cuda.device_of(input1):
            rbot1 = input1.new()
            rbot2 = input2.new()
            output = input1.new()

            correlation_cuda.forward(input1, input2, rbot1, rbot2, output,
                ctx.pad_size, ctx.kernel_size, ctx.max_displacement, ctx.stride1, ctx.stride2, ctx.corr_multiply)

        return output

    @staticmethod
    def backward(ctx, grad_output):
        input1, input2 = ctx.saved_tensors

        with torch.cuda.device_of(input1):
            rbot1 = input1.new()
            rbot2 = input2.new()

            grad_input1 = input1.new()
            grad_input2 = input2.new()

            correlation_cuda.backward(input1, input2, rbot1, rbot2, grad_output, grad_input1, grad_input2,
                ctx.pad_size, ctx.kernel_size, ctx.max_displacement, ctx.stride1, ctx.stride2, ctx.corr_multiply)

        return grad_input1, grad_input2, None, None, None, None, None, None


class Correlation(Module):
    def __init__(self, pad_size=0, kernel_size=0, max_displacement=0, stride1=1, stride2=2, corr_multiply=1):
        super(Correlation, self).__init__()
        self.pad_size = pad_size
        self.kernel_size = kernel_size
        self.max_displacement = max_displacement
        self.stride1 = stride1
        self.stride2 = stride2
        self.corr_multiply = corr_multiply

    def forward(self, input1, input2):

        result = CorrelationFunction.apply(input1, input2, self.pad_size, self.kernel_size, self.max_displacement, self.stride1, self.stride2, self.corr_multiply)

        return result




import argparse

import numpy as np
import torch
from frame_utils import \
    read_gen  # the path is depended on where you create this module
from Networks.FlowNet2 import \
    FlowNet2  # the path is depended on where you create this module

if __name__ == '__main__':
    # obtain the necessary args for construct the flownet framework
    parser = argparse.ArgumentParser()
    parser.add_argument('--fp16', action='store_true', help='Run model in pseudo-fp16 mode (fp16 storage fp32 math).')
    parser.add_argument("--rgb_max", type=float, default=255.)
    
    args = parser.parse_args()

    # initial a Net
    net = FlowNet2(args).cuda()
    # load the state_dict
    dict = torch.load("/home/hjj/PycharmProjects/flownet2_pytorch/FlowNet2_checkpoint.pth.tar")
    net.load_state_dict(dict["state_dict"])

    # load the image pair, you can find this operation in dataset.py
    pim1 = read_gen("/home/hjj/flownet2-master/data/FlyingChairs_examples/0000007-img0.ppm")
    pim2 = read_gen("/home/hjj/flownet2-master/data/FlyingChairs_examples/0000007-img1.ppm")
    images = [pim1, pim2]
    images = np.array(images).transpose(3, 0, 1, 2)
    im = torch.from_numpy(images.astype(np.float32)).unsqueeze(0).cuda()

    # process the image pair to obtian the flow
    result = net(im).squeeze()


    # save flow, I reference the code in scripts/run-flownet.py in flownet2-caffe project
    def writeFlow(name, flow):
        f = open(name, 'wb')
        f.write('PIEH'.encode('utf-8'))
        np.array([flow.shape[1], flow.shape[0]], dtype=np.int32).tofile(f)
        flow = flow.astype(np.float32)
        flow.tofile(f)
        f.flush()
        f.close()


    data = result.data.cpu().numpy().transpose(1, 2, 0)
    writeFlow("/home/hjj/flownet2-master/data/FlyingChairs_examples/0000007-img.flo", data)



import math
import numbers
import random
import sys

import cv2
import numpy as np
import skimage
import skimage.transform

INTER_MODE = {'NEAREST': cv2.INTER_NEAREST, 'BILINEAR': cv2.INTER_LINEAR, 'BICUBIC': cv2.INTER_CUBIC}


def _is_numpy_image(img):
    return isinstance(img, np.ndarray) and (img.ndim in {2, 3})


def rotate(img, angle, resample='BILINEAR', expand=False, center=None):
    imgtype = img.dtype
    if not _is_numpy_image(img):
        raise TypeError('img should be CV Image. Got {}'.format(type(img)))
    h, w, _ = img.shape
    point = center or (w/2, h/2)
    M = cv2.getRotationMatrix2D(point, angle=-angle, scale=1)

    if expand:
        if center is None:
            cos = np.abs(M[0, 0])
            sin = np.abs(M[0, 1])

            # compute the new bounding dimensions of the image
            nW = int((h * sin) + (w * cos))
            nH = int((h * cos) + (w * sin))

            # adjust the rotation matrix to take into account translation
            M[0, 2] += (nW / 2) - point[0]
            M[1, 2] += (nH / 2) - point[1]

            # perform the actual rotation and return the image
            dst = cv2.warpAffine(img, M, (nW, nH))
        else:
            xx = []
            yy = []
            for point in (np.array([0, 0, 1]), np.array([w-1, 0, 1]), np.array([w-1, h-1, 1]), np.array([0, h-1, 1])):
                target = np.matmul(M, point)
                xx.append(target[0])
                yy.append(target[1])
            nh = int(math.ceil(max(yy)) - math.floor(min(yy)))
            nw = int(math.ceil(max(xx)) - math.floor(min(xx)))
            # adjust the rotation matrix to take into account translation
            M[0, 2] += (nw - w)/2
            M[1, 2] += (nh - h)/2
            dst = cv2.warpAffine(img, M, (nw, nh), flags=INTER_MODE[resample])
    else:
        dst = cv2.warpAffine(img, M, (w, h), flags=INTER_MODE[resample])
    return dst.astype(imgtype)


def adjust_brightness(img, brightness_factor):
    if not _is_numpy_image(img):
        raise TypeError('img should be CV Image. Got {}'.format(type(img)))

    im = img.astype(np.float32) * brightness_factor
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_contrast(img, contrast_factor):
    if not _is_numpy_image(img):
        raise TypeError('img should be CV Image. Got {}'.format(type(img)))
    im = img.astype(np.float32)
    mean = round(cv2.cvtColor(im, cv2.COLOR_RGB2GRAY).mean())
    im = (1-contrast_factor)*mean + contrast_factor * im
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_saturation(img, saturation_factor):
    if not _is_numpy_image(img):
        raise TypeError('img should be CV Image. Got {}'.format(type(img)))

    im = img.astype(np.float32)
    degenerate = cv2.cvtColor(cv2.cvtColor(
        im,
        cv2.COLOR_RGB2GRAY), cv2.COLOR_GRAY2RGB)
    im = (1-saturation_factor) * degenerate + saturation_factor * im
    im = im.clip(min=0, max=255)
    return im.astype(img.dtype)


def adjust_hue(img, hue_factor):
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))

    if not _is_numpy_image(img):
        raise TypeError('img should be CV Image. Got {}'.format(type(img)))

    im = img.astype(np.uint8)
    hsv = cv2.cvtColor(im, cv2.COLOR_RGB2HSV_FULL)
    hsv[..., 0] += np.uint8(hue_factor * 255)

    im = cv2.cvtColor(hsv, cv2.COLOR_HSV2RGB_FULL)
    return im.astype(img.dtype)


def flip(img, y_flip=False, x_flip=False, copy=False):
    assert img.ndim == 3, 'The dimension of image must be 3'
    if y_flip:
        img = img[::-1, :, :]
    if x_flip:
        img = img[:, ::-1, :]
    if copy:
        img = img.copy()
    return img


def random_flip_img(img,
                    y_random=True, x_random=True,
                    return_param=False, copy=False):
    y_flip, x_flip = False, False
    if y_random:
        y_flip = random.choice([True, False])
    if x_random:
        x_flip = random.choice([True, False])

    if y_flip:
        img = img[:, ::-1, :]
    if x_flip:
        img = img[:, :, ::-1]

    if copy:
        img = img.copy()

    if return_param:
        return img, {'y_flip': y_flip, 'x_flip': x_flip}
    else:
        return img


def random_flip_imgs(imgs,
                     y_random=True, x_random=True, copy=False):
    if len(imgs) == 0:
        return imgs
    y_flip = False
    if y_random:
        y_flip = random.choice([True, False])
    x_flip = False
    if x_random:
        x_flip = random.choice([True, False])
    return [flip(img, y_flip, x_flip, copy=copy) for img in imgs]


def random_rotate_img(img, degrees):
    if isinstance(degrees, numbers.Number):
        if degrees < 0:
            raise ValueError('If degrees is a single number,'
                             'must be positive')
        degrees = (-degrees, degrees)
    else:
        if len(degrees) != 2:
            raise ValueError('If degrees is a sequence,'
                             'it must be of len 2.')
    angle = random.uniform(degrees[0], degrees[1])
    return skimage.transform.rotate(img, angle)


def random_rotate_imgs(imgs,
                       degrees):
    if isinstance(degrees, numbers.Number):
        if degrees < 0:
            raise ValueError('If degrees is a single number,'
                             'must be positive')
        degrees = (-degrees, degrees)
    else:
        if len(degrees) != 2:
            raise ValueError('If degrees is a sequence,'
                             'it must be of len 2.')
    angle = random.uniform(degrees[0], degrees[1])
    return [skimage.transform.rotate(img, angle) for img in imgs]


def salt_and_pepper(img, prob=0.01):
    imgtype = img.dtype
    rnd = np.random.rand(img.shape[0], img.shape[1])
    noisy = img.copy()
    noisy[rnd < prob/2] = 0.0
    noisy[rnd > 1 - prob/2] = 255.0
    return noisy.astype(imgtype)


class RandomAugmentation(object):

    def __init__(self, degrees,
                 brightness=0.3, contrast=0.3, saturation=0.3, hue=0.3,
                 p=0.5, prob=0.01):
        self.x_flip = random.choice([True, False])

        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError('If degrees is a single number,'
                                 'must be positive')
            degrees = (-degrees, degrees)
        else:
            if len(degrees) != 2:
                raise ValueError('If degrees is a sequence,')
        self.angle = random.uniform(degrees[0], degrees[1])

        transforms = []
        if brightness > 0:
            brightness_factor = random.uniform(max(0, 1 - brightness), 1 + brightness)
            transforms.append(lambda img: adjust_brightness(img, brightness_factor))

        if contrast > 0:
            contrast_factor = random.uniform(max(0, 1 - contrast), 1 + contrast)
            transforms.append(lambda img: adjust_contrast(img, contrast_factor))

        if saturation > 0:
            saturation_factor = random.uniform(max(0, 1 - saturation), 1 + saturation)
            transforms.append(lambda img: adjust_saturation(img, saturation_factor))

        if hue > 0:
            hue_factor = random.uniform(-hue, hue)
            transforms.append(lambda img: adjust_hue(img, hue_factor))

        self.transforms = transforms

        assert isinstance(p, numbers.Number) and p >= 0, 'p should be a positive value'
        assert isinstance(prob, numbers.Number) and prob >= 0, 'p should be a positive value'
        self.p = p
        self.prob = prob

    def __call__(self, img):
        img = img.copy()
        for trans in self.transforms:
            img = trans(img)

        if random.random() < self.p:
            img = salt_and_pepper(img, self.prob)

        img = flip(img, y_flip=False, x_flip=self.x_flip, copy=False)
        img = rotate(img, self.angle)
        img = np.array(np.clip(img, 0, 255),
                 dtype=np.uint8)
        return img


if __name__ == '__main__':
    import os.path as osp

    import cv2

    from anet_tools.video import load_frame

    this_filepath = osp.dirname(__file__)
    video_path = osp.join(this_filepath, 'data', 'motions', 'wiping.mp4')
    frames, stampes = load_frame(video_path)

    while True:
        ra = RandomAugmentation(30)
        key = 0
        for frame in frames:
            frame = ra(frame)
            cv2.imshow('frame', frame)
            key = cv2.waitKey(10)
            if key == ord('q'):
                break
        if key == ord('q'):
            break



import cv2


def load_frame(video_path, start=0.0, duration=-1,
               target_size=None):
    """Load frame

    Parameters
    ----------
    video_path : str or pathlib.Path
        input video path.
    start : float
        start time
    duration : int or float
        duration. If this value is `-1`, load all frames.

    Returns
    -------
    frames : list[numpy.ndarray]
        all frames.
    stamps : list[float]
        time stamps.
    """
    video_path = str(video_path)
    vid = cv2.VideoCapture(video_path)
    vid.set(cv2.CAP_PROP_POS_MSEC, start)
    vid_avail = True
    frames = []
    stamps = []
    while True:
        mills = vid.get(cv2.CAP_PROP_POS_MSEC)
        vid_avail, frame = vid.read()
        if not vid_avail:
            break
        if duration != -1 and mills > start + duration:
            break
        if target_size is not None:
            frame = cv2.resize(frame, target_size)
        frames.append(frame)
        stamps.append(mills / 1000.0)
    vid.release()
    return frames, stamps


def frame_extractor(video_path, start_time, end_time):
    video_path = str(video_path)
    cap = cv2.VideoCapture(video_path)

    fps = cap.get(cv2.CAP_PROP_FPS)
    dt = 1.0 / fps

    second = start_time
    cap.set(cv2.CAP_PROP_POS_MSEC, second * 1000)
    success, image = cap.read()
    while success and second <= end_time:
        second += dt
        yield image
        success, image = cap.read()
    cap.release()



import os.path

import matplotlib.pyplot as plt
import numpy as np

TAG_CHAR = np.array([202021.25], np.float32)

def readFlow(fn):
    """ Read .flo file in Middlebury format"""
    # Code adapted from:
    # http://stackoverflow.com/questions/28013200/reading-middlebury-flow-files-with-python-bytes-array-numpy

    # WARNING: this will work on little-endian architectures (eg Intel x86) only!
    # print 'fn = %s'%(fn)
    with open(fn, 'rb') as f:
        magic = np.fromfile(f, np.float32, count=1)
        if 202021.25 != magic:
            print('Magic number incorrect. Invalid .flo file')
            return None
        else:
            w = np.fromfile(f, np.int32, count=1)
            h = np.fromfile(f, np.int32, count=1)
            # print 'Reading %d x %d flo file\n' % (w, h)
            data = np.fromfile(f, np.float32, count=2*int(w)*int(h))
            # Reshape data into 3D array (columns, rows, bands)
            # The reshape here is for visualization, the original code is (w,h,2)
            return np.resize(data, (int(h), int(w), 2))

def writeFlow(filename,uv,v=None):
    """ Write optical flow to file.
    
    If v is None, uv is assumed to contain both u and v channels,
    stacked in depth.
    Original code by Deqing Sun, adapted from Daniel Scharstein.
    """
    nBands = 2

    if v is None:
        assert(uv.ndim == 3)
        assert(uv.shape[2] == 2)
        u = uv[:,:,0]
        v = uv[:,:,1]
    else:
        u = uv

    assert(u.shape == v.shape)
    height,width = u.shape
    f = open(filename,'wb')
    # write the header
    f.write(TAG_CHAR)
    np.array(width).astype(np.int32).tofile(f)
    np.array(height).astype(np.int32).tofile(f)
    # arrange into matrix form
    tmp = np.zeros((height, width*nBands))
    tmp[:,np.arange(width)*2] = u
    tmp[:,np.arange(width)*2 + 1] = v
    tmp.astype(np.float32).tofile(f)
    f.close()


# ref: https://github.com/sampepose/flownet2-tf/
# blob/18f87081db44939414fc4a48834f9e0da3e69f4c/src/flowlib.py#L240
def visulize_flow_file(flow_filename, save_dir=None):
	flow_data = readFlow(flow_filename)
	img = flow2img(flow_data)
	# plt.imshow(img)
	# plt.show()
	if save_dir:
		idx = flow_filename.rfind("/") + 1
		plt.imsave(os.path.join(save_dir, "%s-vis.png" % flow_filename[idx:-4]), img)


def flow2img(flow_data):
	"""
	convert optical flow into color image
	:param flow_data:
	:return: color image
	"""
	# print(flow_data.shape)
	# print(type(flow_data))
	u = flow_data[:, :, 0]
	v = flow_data[:, :, 1]

	UNKNOW_FLOW_THRESHOLD = 1e7
	pr1 = abs(u) > UNKNOW_FLOW_THRESHOLD
	pr2 = abs(v) > UNKNOW_FLOW_THRESHOLD
	idx_unknown = (pr1 | pr2)
	u[idx_unknown] = v[idx_unknown] = 0

	# get max value in each direction
	maxu = -999.
	maxv = -999.
	minu = 999.
	minv = 999.
	maxu = max(maxu, np.max(u))
	maxv = max(maxv, np.max(v))
	minu = min(minu, np.min(u))
	minv = min(minv, np.min(v))

	rad = np.sqrt(u ** 2 + v ** 2)
	maxrad = max(-1, np.max(rad))
	u = u / maxrad + np.finfo(float).eps
	v = v / maxrad + np.finfo(float).eps

	img = compute_color(u, v)

	idx = np.repeat(idx_unknown[:, :, np.newaxis], 3, axis=2)
	img[idx] = 0

	return np.uint8(img)


def compute_color(u, v):
	"""
	compute optical flow color map
	:param u: horizontal optical flow
	:param v: vertical optical flow
	:return:
	"""

	height, width = u.shape
	img = np.zeros((height, width, 3))

	NAN_idx = np.isnan(u) | np.isnan(v)
	u[NAN_idx] = v[NAN_idx] = 0

	colorwheel = make_color_wheel()
	ncols = np.size(colorwheel, 0)

	rad = np.sqrt(u ** 2 + v ** 2)

	a = np.arctan2(-v, -u) / np.pi

	fk = (a + 1) / 2 * (ncols - 1) + 1

	k0 = np.floor(fk).astype(int)

	k1 = k0 + 1
	k1[k1 == ncols + 1] = 1
	f = fk - k0

	for i in range(0, np.size(colorwheel, 1)):
		tmp = colorwheel[:, i]
		col0 = tmp[k0 - 1] / 255
		col1 = tmp[k1 - 1] / 255
		col = (1 - f) * col0 + f * col1

		idx = rad <= 1
		col[idx] = 1 - rad[idx] * (1 - col[idx])
		notidx = np.logical_not(idx)

		col[notidx] *= 0.75
		img[:, :, i] = np.uint8(np.floor(255 * col * (1 - NAN_idx)))

	return img


def make_color_wheel():
	"""
	Generate color wheel according Middlebury color code
	:return: Color wheel
	"""
	RY = 15
	YG = 6
	GC = 4
	CB = 11
	BM = 13
	MR = 6

	ncols = RY + YG + GC + CB + BM + MR

	colorwheel = np.zeros([ncols, 3])

	col = 0

	# RY
	colorwheel[0:RY, 0] = 255
	colorwheel[0:RY, 1] = np.transpose(np.floor(255 * np.arange(0, RY) / RY))
	col += RY

	# YG
	colorwheel[col:col + YG, 0] = 255 - np.transpose(np.floor(255 * np.arange(0, YG) / YG))
	colorwheel[col:col + YG, 1] = 255
	col += YG

	# GC
	colorwheel[col:col + GC, 1] = 255
	colorwheel[col:col + GC, 2] = np.transpose(np.floor(255 * np.arange(0, GC) / GC))
	col += GC

	# CB
	colorwheel[col:col + CB, 1] = 255 - np.transpose(np.floor(255 * np.arange(0, CB) / CB))
	colorwheel[col:col + CB, 2] = 255
	col += CB

	# BM
	colorwheel[col:col + BM, 2] = 255
	colorwheel[col:col + BM, 0] = np.transpose(np.floor(255 * np.arange(0, BM) / BM))
	col += + BM

	# MR
	colorwheel[col:col + MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
	colorwheel[col:col + MR, 0] = 255

	return colorwheel



# flake8: noqa

from anet_tools.video.load_frame import load_frame
from anet_tools.video.video_info import get_video_count, get_video_duration
from anet_tools.video.video_proc import ImageSequenceProc, VideoProc



import json
import os
import os.path as osp

from easydict import EasyDict

"""
This file contains the setting for running ActivityNet related experiments.
"""

ANET_CFG = EasyDict()

# Version and other macro settings

cur_dir = osp.dirname(__file__)

ANET_DB_VERSIONS = {
    '1.2': osp.join(cur_dir, 'data/activity_net.v1-2.json'),
    '1.3': osp.join(cur_dir, 'data/activity_net.v1-3.json'),
}

# Force the leaf node to be included in the label list
ANET_CFG.FORCE_INCLUDE = {"1.3": [], "1.2": []}


# Allow using external config files to override the above settings
def LoadExternalYAMLConfig(yaml_file):
    import yaml
    new_cfg = yaml.load(open(yaml_file))
    ANET_CFG.update(new_cfg)


import glob


def get_all_media_files(src_folders, accepted_extensions):
    media_files = []
    for f in src_folders:
        all_files = glob.glob(os.path.join(f, "*"))
        media_files.extend([name for name in all_files if os.path.splitext(name)[-1] in accepted_extensions])
    return media_files


import logging
import sys

_formatter = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - [%(levelname)s] %(message)s')
_ch = logging.StreamHandler()
_ch.setLevel(logging.DEBUG)
_ch.setFormatter(_formatter)

def get_logger(debug=False):
    logger = logging.getLogger("pyActionRec")
    logger.setLevel(logging.DEBUG if debug else logging.INFO)
    logger.addHandler(_ch)
    return logger


logger = get_logger()


class Instance(object):
    """
    Representing an instance of activity in the videos
    """

    def __init__(self, idx, anno, vid_id, vid_info, name_num_mapping):
        self._starting, self._ending = anno['segment'][0], anno['segment'][1]
        self._str_label = anno['label']
        self._total_duration = vid_info['duration']
        self._idx = idx
        self._vid_id = vid_id
        self._file_path = None

        if name_num_mapping:
            self._num_label = name_num_mapping[self._str_label]

    @property
    def start_time(self):
        return self._starting

    @property
    def end_time(self):
        return self._ending

    @property
    def time_span(self):
        return self._starting, self._ending

    @property
    def covering_ratio(self):
        return self._starting / float(self._total_duration), self._ending / float(self._total_duration)

    @property
    def num_label(self):
        return self._num_label

    @property
    def label(self):
        return self._str_label

    @property
    def name(self):
        return '{}_{}'.format(self._vid_id, self._idx)

    @property
    def path(self):
        if self._file_path is None:
            raise ValueError("This instance is not associated to a file on disk. Maybe the file is missing?")
        return self._file_path

    @path.setter
    def path(self, path):
        self._file_path = path


class Video(object):
    """
    This class represents one video in the activity-net db
    """
    def __init__(self, key, info, name_idx_mapping=None):
        self._id = key
        self._info_dict = info
        self._instances = [Instance(i, x, self._id, self._info_dict, name_idx_mapping)
                           for i, x in enumerate(self._info_dict['annotations'])]
        self._file_path = None

    @property
    def id(self):
        return self._id

    @property
    def url(self):
        return self._info_dict['url']

    @property
    def instances(self):
        return self._instances

    @property
    def duration(self):
        return self._info_dict['duration']

    @property
    def subset(self):
        return self._info_dict['subset']

    @property
    def instance(self):
        return self._instances

    @property
    def path(self):
        if self._file_path is None:
            raise ValueError("This video is not associated to a file on disk. Maybe the file is missing?")
        return self._file_path

    @path.setter
    def path(self, path):
        self._file_path = path


class ANetDB(object):
    """
    This class is the abstraction of the activity-net db
    """

    def __init__(self, version="1.3"):
        if version not in ANET_DB_VERSIONS:
            raise ValueError("Unsupported database version {}".format(version))

        raw_db_file = os.path.join(ANET_DB_VERSIONS[version])
        db_data = json.load(open(raw_db_file))

        self.version = version
        self._prepare_data(db_data)

    def _prepare_data(self, raw_db):
        self._version = raw_db['version']

        # deal with taxonomy
        self._taxonomy = raw_db['taxonomy']
        self._parse_taxonomy()

        self._database = raw_db['database']
        self._video_dict = {k: Video(k, v, self._name_idx_table) for k,v in self._database.items()}

        # split testing/training/validation set
        self._testing_dict = {k: v for k, v in self._video_dict.items() if v.subset == 'testing'}
        self._training_dict = {k: v for k, v in self._video_dict.items() if v.subset == 'training'}
        self._validation_dict = {k: v for k, v in self._video_dict.items() if v.subset == 'validation'}

        self._training_inst_dict = {i.name: i for v in self._training_dict.values() for i in v.instances}
        self._validation_inst_dict = {i.name: i for v in self._validation_dict.values() for i in v.instances}

    def get_ordered_label_list(self):
        """
        This function return a list of textual labels corresponding the numerical labels. For example, if one samples is
        classified to class 18, the textual label is the 18-th element in the returned list.
        Returns: list
        """
        return [self._idx_name_table[x] for x in sorted(self._idx_name_table.keys())]

    def get_subset_videos(self, subset_name):
        if subset_name == 'training':
            return self._training_dict.values()
        elif subset_name == 'validation':
            return self._validation_dict.values()
        elif subset_name == 'testing':
            return self._testing_dict.values()
        else:
            raise ValueError("Unknown subset {}".format(subset_name))

    def get_subset_instance(self, subset_name):
        if subset_name == 'training':
            return self._training_inst_dict.values()
        elif subset_name == 'validation':
            return self._validation_inst_dict.values()
        else:
            raise ValueError("Unknown subset {}".format(subset_name))

    def _parse_taxonomy(self):
        """
        This function just parses the taxonomy file
        It gives alphabetical ordered indices to the classes in competition
        :return:
        """
        name_dict = {x['nodeName']: x for x in self._taxonomy}
        parents = set()
        for x in self._taxonomy:
            parents.add(x['parentName'])

        # leaf nodes are those without any child
        leaf_nodes = [name_dict[x] for x
                      in list(set(name_dict.keys()).difference(parents)) + ANET_CFG.FORCE_INCLUDE.setdefault(self.version, [])]
        sorted_lead_nodes = sorted(leaf_nodes, key=lambda l: l['nodeName'])
        self._idx_name_table = {i: e['nodeName'] for i, e in enumerate(sorted_lead_nodes)}
        self._name_idx_table = {e['nodeName']: i for i, e in enumerate(sorted_lead_nodes)}
        self._name_table = {x['nodeName']: x for x in sorted_lead_nodes}

    def parse_activitynet_splits(self):
        train_instance = self.get_subset_instance('training')
        val_instance = self.get_subset_instance('validation')
        test_instance = self.get_subset_videos('testing')

        splits = []

        train_list = [(x.name, x.num_label) for x in train_instance]
        val_list = [(x.name, x.num_label) for x in val_instance]
        test_list = [(x.id, 0) for x in test_instance]

        splits.append((train_list, val_list))
        splits.append((train_list + val_list, test_list))

        return splits


if __name__ == '__main__':
    db = ANetDB()
    logger.info("Database construction success".format())
